import json
import urllib.parse
from typing import List, Dict

import pytest
import sqlalchemy as sql

import pixeltable as pxt
import pixeltable.func as func
from pixeltable import catalog
from pixeltable import exceptions as excs
from pixeltable import exprs
from pixeltable.exprs import Expr, ColumnRef
from pixeltable.exprs import RELATIVE_PATH_ROOT as R
from pixeltable.functions import cast, sum, count
from pixeltable.functions.pil.image import blend
from pixeltable.iterators import FrameIterator
from pixeltable.tests.utils import get_image_files, skip_test_if_not_installed
from pixeltable.type_system import StringType, BoolType, IntType, ArrayType, ColumnType, FloatType, \
    VideoType


class TestExprs:
    def test_basic(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        assert t['c1'].equals(t.c1)
        assert t['c7']['*'].f5.equals(t.c7['*'].f5)

        assert isinstance(t.c1 == None, Expr)
        assert isinstance(t.c1 < 'a', Expr)
        assert isinstance(t.c1 <= 'a', Expr)
        assert isinstance(t.c1 == 'a', Expr)
        assert isinstance(t.c1 != 'a', Expr)
        assert isinstance(t.c1 > 'a', Expr)
        assert isinstance(t.c1 >= 'a', Expr)
        assert isinstance((t.c1 == 'a') & (t.c2 < 5), Expr)
        assert isinstance((t.c1 == 'a') | (t.c2 < 5), Expr)
        assert isinstance(~(t.c1 == 'a'), Expr)
        with pytest.raises(AttributeError) as excinfo:
            _ = t.does_not_exist
        assert 'unknown' in str(excinfo.value).lower()

    def test_compound_predicates(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        # compound predicates that can be fully evaluated in SQL
        _ = t.where((t.c1 == 'test string') & (t.c6.f1 > 50)).collect()
        _ = t.where((t.c1 == 'test string') & (t.c2 > 50)).collect()
        e = ((t.c1 == 'test string') & (t.c2 > 50)).sql_expr()
        assert len(e.clauses) == 2

        e = ((t.c1 == 'test string') & (t.c2 > 50) & (t.c3 < 1.0)).sql_expr()
        assert len(e.clauses) == 3
        e = ((t.c1 == 'test string') | (t.c2 > 50)).sql_expr()
        assert len(e.clauses) == 2
        e = ((t.c1 == 'test string') | (t.c2 > 50) | (t.c3 < 1.0)).sql_expr()
        assert len(e.clauses) == 3
        e = (~(t.c1 == 'test string')).sql_expr()
        assert isinstance(e, sql.sql.expression.BinaryExpression)

        with pytest.raises(TypeError) as exc_info:
            _ = t.where((t.c1 == 'test string') or (t.c6.f1 > 50)).collect()
        assert 'cannot be used in conjunction with python boolean operators' in str(exc_info.value).lower()

        # compound predicates with Python functions
        @pxt.udf(return_type=BoolType(), param_types=[StringType()])
        def udf(_: str) -> bool:
            return True
        @pxt.udf(return_type=BoolType(), param_types=[IntType()])
        def udf2(_: int) -> bool:
            return True

        # TODO: find a way to test this
        # # & can be split
        # p = (t.c1 == 'test string') & udf(t.c1)
        # assert p.sql_expr() is None
        # sql_pred, other_pred = p.extract_sql_predicate()
        # assert isinstance(sql_pred, sql.sql.expression.BinaryExpression)
        # assert isinstance(other_pred, FunctionCall)
        #
        # p = (t.c1 == 'test string') & udf(t.c1) & (t.c2 > 50)
        # assert p.sql_expr() is None
        # sql_pred, other_pred = p.extract_sql_predicate()
        # assert len(sql_pred.clauses) == 2
        # assert isinstance(other_pred, FunctionCall)
        #
        # p = (t.c1 == 'test string') & udf(t.c1) & (t.c2 > 50) & udf2(t.c2)
        # assert p.sql_expr() is None
        # sql_pred, other_pred = p.extract_sql_predicate()
        # assert len(sql_pred.clauses) == 2
        # assert isinstance(other_pred, CompoundPredicate)
        #
        # # | cannot be split
        # p = (t.c1 == 'test string') | udf(t.c1)
        # assert p.sql_expr() is None
        # sql_pred, other_pred = p.extract_sql_predicate()
        # assert sql_pred is None
        # assert isinstance(other_pred, CompoundPredicate)

    def test_filters(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        _ = t[t.c1 == 'test string'].show()
        print(_)
        _ = t[t.c2 > 50].show()
        print(_)
        _ = t[t.c1n == None].show()
        print(_)
        _ = t[t.c1n != None].show(0)
        print(_)

    def test_exception_handling(self, test_tbl: catalog.Table) -> None:
        t = test_tbl

        # error in expr that's handled in SQL
        with pytest.raises(excs.Error):
            _ = t[(t.c2 + 1) / t.c2].show()

        # error in expr that's handled in Python
        with pytest.raises(excs.Error):
            _ = t[(t.c6.f2 + 1) / (t.c2 - 10)].show()

        # the same, but with an inline function
        @pxt.udf(return_type=FloatType(), param_types=[IntType(), IntType()])
        def f(a: int, b: int) -> float:
            return a / b
        with pytest.raises(excs.Error):
            _ = t[f(t.c2 + 1, t.c2)].show()

        # error in agg.init()
        @pxt.uda(update_types=[IntType()], value_type=IntType(), name='agg')
        class Aggregator(pxt.Aggregator):
            def __init__(self):
                self.sum = 1 / 0
            def update(self, val):
                pass
            def value(self):
                return 1
        with pytest.raises(excs.Error):
            _ = t[agg(t.c2)].show()

        # error in agg.update()
        @pxt.uda(update_types=[IntType()], value_type=IntType(), name='agg')
        class Aggregator(pxt.Aggregator):
            def __init__(self):
                self.sum = 0
            def update(self, val):
                self.sum += 1 / val
            def value(self):
                return 1
        with pytest.raises(excs.Error):
            _ = t[agg(t.c2 - 10)].show()

        # error in agg.value()
        @pxt.uda(update_types=[IntType()], value_type=IntType(), name='agg')
        class Aggregator(pxt.Aggregator):
            def __init__(self):
                self.sum = 0
            def update(self, val):
                self.sum += val
            def value(self):
                return 1 / self.sum
        with pytest.raises(excs.Error):
            _ = t[t.c2 <= 2][agg(t.c2 - 1)].show()

    def test_props(self, test_tbl: catalog.Table, img_tbl: catalog.Table) -> None:
        t = test_tbl
        # errortype/-msg for computed column
        res = t.select(error=t.c8.errortype).collect()
        assert res.to_pandas()['error'].isna().all()
        res = t.select(error=t.c8.errormsg).collect()
        assert res.to_pandas()['error'].isna().all()

        img_t = img_tbl
        # fileurl
        res = img_t.select(img_t.img.fileurl).show(0).to_pandas()
        stored_urls = set(res.iloc[:, 0])
        assert len(stored_urls) == len(res)
        all_urls  = set([urllib.parse.urljoin('file:', path) for path in get_image_files()])
        assert stored_urls <= all_urls

        # localpath
        res = img_t.select(img_t.img.localpath).show(0).to_pandas()
        stored_paths = set(res.iloc[:, 0])
        assert len(stored_paths) == len(res)
        all_paths  = set(get_image_files())
        assert stored_paths <= all_paths

        # errortype/-msg for image column
        res = img_t.select(error=img_t.img.errortype).collect().to_pandas()
        assert res['error'].isna().all()
        res = img_t.select(error=img_t.img.errormsg).collect().to_pandas()
        assert res['error'].isna().all()

        for c in [t.c1, t.c1n, t.c2, t.c3, t.c4, t.c5, t.c6, t.c7]:
            # errortype/errormsg only applies to stored computed and media columns
            with pytest.raises(excs.Error) as excinfo:
                _ = t.select(c.errortype).show()
            assert 'only valid for' in str(excinfo.value)
            with pytest.raises(excs.Error) as excinfo:
                _ = t.select(c.errormsg).show()
            assert 'only valid for' in str(excinfo.value)

            # fileurl/localpath only applies to media columns
            with pytest.raises(excs.Error) as excinfo:
                _ = t.select(t.c1.fileurl).show()
            assert 'only valid for' in str(excinfo.value)
            with pytest.raises(excs.Error) as excinfo:
                _ = t.select(t.c1.localpath).show()
            assert 'only valid for' in str(excinfo.value)

        # fileurl/localpath doesn't apply to unstored computed img columns
        img_t.add_column(c9=img_t.img.rotate(30))
        with pytest.raises(excs.Error) as excinfo:
            _ = img_t.select(img_t.c9.localpath).show()
        assert 'computed unstored' in str(excinfo.value)

    def test_null_args(self, test_client: pxt.Client) -> None:
        # create table with two int columns
        schema = {'c1': FloatType(nullable=True), 'c2': FloatType(nullable=True)}
        t = test_client.create_table('test', schema)

        # computed column that doesn't allow nulls
        t.add_column(c3=lambda c1, c2: c1 + c2, type=FloatType(nullable=False))
        # function that does allow nulls
        @pxt.udf(return_type=FloatType(nullable=True),
                 param_types=[FloatType(nullable=False), FloatType(nullable=True)])
        def f(a: int, b: int) -> int:
            if b is None:
                return a
            return a + b
        t.add_column(c4=f(t.c1, t.c2))

        # data that tests all combinations of nulls
        data = [{'c1': 1.0, 'c2': 1.0}, {'c1': 1.0, 'c2': None}, {'c1': None, 'c2': 1.0}, {'c1': None, 'c2': None}]
        status = t.insert(data, fail_on_exception=False)
        assert status.num_rows == len(data)
        assert status.num_excs == len(data) - 1
        result = t.select(t.c3, t.c4).collect()
        assert result['c3'] == [2.0, None, None, None]
        assert result['c4'] == [2.0, 1.0, None, None]

    def test_arithmetic_exprs(self, test_tbl: catalog.Table) -> None:
        t = test_tbl

        _ = t[t.c2, t.c6.f3, t.c2 + t.c6.f3, (t.c2 + t.c6.f3) / (t.c6.f3 + 1)].show()
        _ = t[t.c2 + t.c2].show()
        for op1, op2 in [(t.c2, t.c2), (t.c3, t.c3)]:
            _ = t[op1 + op2].show()
            _ = t[op1 - op2].show()
            _ = t[op1 * op2].show()
            _ = t[op1 > 0][op1 / op2].show()

        # non-numeric types
        for op1, op2 in [
            (t.c1, t.c2), (t.c1, 1), (t.c2, t.c1), (t.c2, 'a'),
            (t.c1, t.c3), (t.c1, 1.0), (t.c3, t.c1), (t.c3, 'a')
        ]:
            with pytest.raises(excs.Error):
                _ = t[op1 + op2]
            with pytest.raises(excs.Error):
                _ = t[op1 - op2]
            with pytest.raises(excs.Error):
                _ = t[op1 * op2]
            with pytest.raises(excs.Error):
                _ = t[op1 / op2]

        # TODO: test division; requires predicate
        for op1, op2 in [(t.c6.f2, t.c6.f2), (t.c6.f3, t.c6.f3)]:
            _ = t[op1 + op2].show()
            _ = t[op1 - op2].show()
            _ = t[op1 * op2].show()
            with pytest.raises(excs.Error):
                _ = t[op1 / op2].show()

        for op1, op2 in [
            (t.c6.f1, t.c6.f2), (t.c6.f1, t.c6.f3), (t.c6.f1, 1), (t.c6.f1, 1.0),
            (t.c6.f2, t.c6.f1), (t.c6.f3, t.c6.f1), (t.c6.f2, 'a'), (t.c6.f3, 'a'),
        ]:
            with pytest.raises(excs.Error):
                _ = t[op1 + op2].show()
            with pytest.raises(excs.Error):
                _ = t[op1 - op2].show()
            with pytest.raises(excs.Error):
                _ = t[op1 * op2].show()


    def test_inline_dict(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        df = t[[{'a': t.c1, 'b': {'c': t.c2}, 'd': 1, 'e': {'f': 2}}]]
        result = df.show()
        print(result)

    def test_inline_array(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        result = t.select([[t.c2, 1], [t.c2, 2]]).show()
        t = result.column_types()[0]
        assert t.is_array_type()
        assert isinstance(t, ArrayType)
        assert t.shape == (2, 2)
        assert t.dtype == ColumnType.Type.INT

    def test_json_mapper(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        # top-level is dict
        df = t[t.c6.f5['*'] >> (R + 1)]
        res = df.show()
        print(res)
        _ = t[t.c7['*'].f5 >> [R[3], R[2], R[1], R[0]]]
        _ = _.show()
        print(_)
        # target expr contains global-scope dependency
        df = t[
            t.c6.f5['*'] >> (R * t.c6.f5[1])
        ]
        res = df.show()
        print(res)

    def test_dicts(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        # top-level is dict
        _ = t[t.c6.f1]
        _ = _.show()
        print(_)
        # predicate on dict field
        _ = t[t.c6.f2 < 2].show()
        #_ = t[t.c6.f2].show()
        #_ = t[t.c6.f5].show()
        _ = t[t.c6.f6.f8].show()
        _ = t[cast(t.c6.f6.f8, ArrayType((4,), FloatType()))].show()

        # top-level is array
        #_ = t[t.c7['*'].f1].show()
        #_ = t[t.c7['*'].f2].show()
        #_ = t[t.c7['*'].f5].show()
        _ = t[t.c7['*'].f6.f8].show()
        _ = t[t.c7[0].f6.f8].show()
        _ = t[t.c7[:2].f6.f8].show()
        _ = t[t.c7[::-1].f6.f8].show()
        _ = t[cast(t.c7['*'].f6.f8, ArrayType((2, 4), FloatType()))].show()
        print(_)

    def test_arrays(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        t.add_column(array_col=[[t.c2, 1], [1, t.c2]])
        _ = t[t.array_col].show()
        print(_)
        _ = t[t.array_col[:, 0]].show()
        print(_)

    def test_astype(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        # Convert int to float
        status = t.add_column(c2_as_float=t.c2.astype(FloatType()))
        assert status.num_excs == 0
        data = t.select(t.c2, t.c2_as_float).collect()
        for row in data:
            assert isinstance(row['c2'], int)
            assert isinstance(row['c2_as_float'], float)
            assert row['c2'] == row['c2_as_float']
        # Compound expression
        status = t.add_column(compound_as_float=(t.c2 + 1).astype(FloatType()))
        assert status.num_excs == 0
        data = t.select(t.c2, t.compound_as_float).collect()
        for row in data:
            assert isinstance(row['compound_as_float'], float)
            assert row['c2'] + 1 == row['compound_as_float']
        # Type conversion error
        status = t.add_column(c2_as_string=t.c2.astype(StringType()))
        assert status.num_excs == t.count()

    def test_apply(self, test_tbl: catalog.Table) -> None:

        t = test_tbl

        # For each column c1, ..., c5, we create a new column ci_as_str that converts it to
        # a string, then check that each row is correctly converted
        # (For c1 this is the no-op string-to-string conversion)
        for col_id in range(1, 6):
            col_name = f'c{col_id}'
            str_col_name = f'c{col_id}_str'
            status = t.add_column(**{str_col_name: t[col_name].apply(str)})
            assert status.num_excs == 0
            data = t.select(t[col_name], t[str_col_name]).collect()
            for row in data:
                assert row[str_col_name] == str(row[col_name])

        # Test a compound expression with apply
        status = t.add_column(c2_plus_1_str=(t.c2 + 1).apply(str))
        assert status.num_excs == 0
        data = t.select(t.c2, t.c2_plus_1_str).collect()
        for row in data:
            assert row['c2_plus_1_str'] == str(row['c2'] + 1)

        # For columns c6, c7, try using json.dumps and json.loads to emit and parse JSON <-> str
        for col_id in range(6, 8):
            col_name = f'c{col_id}'
            str_col_name = f'c{col_id}_str'
            back_to_json_col_name = f'c{col_id}_back_to_json'
            status = t.add_column(**{str_col_name: t[col_name].apply(json.dumps)})
            assert status.num_excs == 0
            status = t.add_column(**{back_to_json_col_name: t[str_col_name].apply(json.loads)})
            assert status.num_excs == 0
            data = t.select(t[col_name], t[str_col_name], t[back_to_json_col_name]).collect()
            for row in data:
                assert row[str_col_name] == json.dumps(row[col_name])
                assert row[back_to_json_col_name] == row[col_name]

        def f1(x):
            return str(x)

        # Now test that a function without a return type throws an exception ...
        with pytest.raises(excs.Error) as exc_info:
            t.c2.apply(f1)
        assert 'Column type of `f1` cannot be inferred.' in str(exc_info.value)

        # ... but works if the type is specified explicitly.
        status = t.add_column(c2_str_f1=t.c2.apply(f1, col_type=StringType()))
        assert status.num_excs == 0

        # Test that the return type of a function can be successfully inferred.
        def f2(x) -> str:
            return str(x)

        status = t.add_column(c2_str_f2=t.c2.apply(f2))
        assert status.num_excs == 0

        # Test various validation failures.

        def f3(x, y) -> str:
            return f'{x}{y}'

        with pytest.raises(excs.Error) as exc_info:
            t.c2.apply(f3)  # Too many required parameters
        assert str(exc_info.value) == 'Function `f3` has multiple required parameters.'

        def f4() -> str:
            return "pixeltable"

        with pytest.raises(excs.Error) as exc_info:
            t.c2.apply(f4)  # No positional parameters
        assert str(exc_info.value) == 'Function `f4` has no positional parameters.'

        def f5(**kwargs) -> str:
            return ""

        with pytest.raises(excs.Error) as exc_info:
            t.c2.apply(f5)  # No positional parameters
        assert str(exc_info.value) == 'Function `f5` has no positional parameters.'

        # Ensure these varargs signatures are acceptable

        def f6(x, **kwargs) -> str:
            return x

        t.c2.apply(f6)

        def f7(x, *args) -> str:
            return x

        t.c2.apply(f7)

        def f8(*args) -> str:
            return ''

        t.c2.apply(f8)

    def test_select_list(self, img_tbl) -> None:
        t = img_tbl
        result = t[t.img].show(n=100)
        _ = result._repr_html_()
        df = t[[t.img, t.img.rotate(60)]]
        _ = df.show(n=100)._repr_html_()

        with pytest.raises(excs.Error):
            _ = t[t.img.rotate]

    def test_img_members(self, img_tbl) -> None:
        t = img_tbl
        # make sure the limit is applied in Python, not in the SELECT
        result = t[t.img.height > 200][t.img].show(n=3)
        assert len(result) == 3
        result = t[t.img.crop((10, 10, 60, 60))].show(n=100)
        result = t[t.img.crop((10, 10, 60, 60)).resize((100, 100))].show(n=100)
        result = t[t.img.crop((10, 10, 60, 60)).resize((100, 100)).convert('L')].show(n=100)
        result = t[t.img.getextrema()].show(n=100)
        result = t[t.img, t.img.height, t.img.rotate(90)].show(n=100)
        _ = result._repr_html_()

    def test_img_functions(self, img_tbl) -> None:
        skip_test_if_not_installed('nos')
        t = img_tbl
        from pixeltable.functions.pil.image import resize
        result = t[t.img.resize((224, 224))].show(0)
        result = t[resize(t.img, (224, 224))].show(0)
        result = t[blend(t.img, t.img.rotate(90), 0.5)].show(100)
        print(result)
        from pixeltable.functions.nos.image_embedding import openai_clip
        result = t[openai_clip(t.img.resize((224, 224)))].show(10)
        print(result)
        _ = result._repr_html_()
        _ = t.img.entropy() > 1
        _ = (t.img.entropy() > 1) & (t.split == 'train')
        _ = (t.img.entropy() > 1) & (t.split == 'train') & (t.split == 'val')
        _ = (t.split == 'train') & (t.img.entropy() > 1) & (t.split == 'val') & (t.img.entropy() < 0)
        _ = t[(t.split == 'train') & (t.category == 'n03445777')][t.img].show()
        print(_)
        result = t[t.img.width > 1].show()
        print(result)
        result = t[(t.split == 'val') & (t.img.entropy() > 1) & (t.category == 'n03445777')].show()
        print(result)
        result = t[
            (t.split == 'train') & (t.img.entropy() > 1) & (t.split == 'val') & (t.img.entropy() < 0)
        ][t.img, t.split].show()
        print(result)

    def test_similarity(self, indexed_img_tbl: catalog.Table) -> None:
        skip_test_if_not_installed('nos')
        t = indexed_img_tbl
        _ = t.show(30)
        probe = t.select(t.img, t.category).show(1)
        img = probe[0, 0]
        result = t.where(t.img.nearest(img)).show(10)
        assert len(result) == 10
        # nearest() with one SQL predicate and one Python predicate
        result = t[t.img.nearest(img) & (t.category == probe[0, 1]) & (t.img.width > 1)].show(10)
        # TODO: figure out how to verify results

        with pytest.raises(excs.Error) as exc_info:
            _ = t[t.img.nearest(img)].order_by(t.category).show()
        assert 'cannot be used in conjunction with' in str(exc_info.value)

        result = t[t.img.nearest('musical instrument')].show(10)
        assert len(result) == 10
        # matches() with one SQL predicate and one Python predicate
        french_horn_category = 'n03394916'
        result = t[
            t.img.nearest('musical instrument') & (t.category == french_horn_category) & (t.img.width > 1)
        ].show(10)

        with pytest.raises(excs.Error) as exc_info:
            _ = t[t.img.nearest(5)].show()
        assert 'requires' in str(exc_info.value)

    # TODO: this doesn't work when combined with test_similarity(), for some reason the data table for img_tbl
    # doesn't get created; why?
    def test_similarity2(self, img_tbl: catalog.Table) -> None:
        t = img_tbl
        probe = t[t.img].show(1)
        img = probe[0, 0]

        with pytest.raises(excs.Error):
            _ = t[t.img.nearest(img)].show(10)
        with pytest.raises(excs.Error):
            _ = t[t.img.nearest('musical instrument')].show(10)

    def test_ids(
            self, test_tbl: catalog.Table, test_tbl_exprs: List[exprs.Expr],
            img_tbl: catalog.Table, img_tbl_exprs: List[exprs.Expr]
    ) -> None:
        d: Dict[int, exprs.Expr] = {}
        for e in test_tbl_exprs:
            assert e.id is not None
            d[e.id] = e
        for e in img_tbl_exprs:
            assert e.id is not None
            d[e.id] = e
        assert len(d) == len(test_tbl_exprs) + len(img_tbl_exprs)

    def test_serialization(
            self, test_tbl_exprs: List[exprs.Expr], img_tbl_exprs: List[exprs.Expr]
    ) -> None:
        """Test as_dict()/from_dict() (via serialize()/deserialize()) for all exprs."""
        for e in test_tbl_exprs:
            e_serialized = e.serialize()
            e_deserialized = Expr.deserialize(e_serialized)
            assert e.equals(e_deserialized)

        for e in img_tbl_exprs:
            e_serialized = e.serialize()
            e_deserialized = Expr.deserialize(e_serialized)
            assert e.equals(e_deserialized)

    def test_print(self, test_tbl_exprs: List[exprs.Expr], img_tbl_exprs: List[exprs.Expr]) -> None:
        _ = func.FunctionRegistry.get().module_fns
        for e in test_tbl_exprs:
            _ = str(e)
            print(_)
        for e in img_tbl_exprs:
            _ = str(e)
            print(_)

    def test_subexprs(self, img_tbl: catalog.Table) -> None:
        t = img_tbl
        e = t.img
        subexprs = [s for s in e.subexprs()]
        assert len(subexprs) == 1
        e = t.img.rotate(90).resize((224, 224))
        subexprs = [s for s in e.subexprs()]
        assert len(subexprs) == 4
        subexprs = [s for s in e.subexprs(expr_class=ColumnRef)]
        assert len(subexprs) == 1
        assert t.img.equals(subexprs[0])

    def test_window_fns(self, test_client: pxt.Client, test_tbl: catalog.Table) -> None:
        cl = test_client
        t = test_tbl
        _ = t.select(sum(t.c2, group_by=t.c4, order_by=t.c3)).show(100)

        # conflicting ordering requirements
        with pytest.raises(excs.Error):
            _ = t.select(sum(t.c2, group_by=t.c4, order_by=t.c3), sum(t.c2, group_by=t.c3, order_by=t.c4)).show(100)
        with pytest.raises(excs.Error):
            _ = t.select(sum(t.c2, group_by=t.c4, order_by=t.c3), sum(t.c2, group_by=t.c3, order_by=t.c4)).show(100)

        # backfill works
        t.add_column(c9=sum(t.c2, group_by=t.c4, order_by=t.c3))
        _ = t.c9.col.has_window_fn_call()

        # ordering conflict between frame extraction and window fn
        base_t = cl.create_table('videos', {'video': VideoType(), 'c2': IntType(nullable=False)})
        args = {'video': base_t.video, 'fps': 0}
        v = cl.create_view('frame_view', base_t, iterator_class=FrameIterator, iterator_args=args)
        # compatible ordering
        _ = v.select(v.frame, sum(v.frame_idx, group_by=base_t, order_by=v.pos)).show(100)
        with pytest.raises(excs.Error):
            # incompatible ordering
            _ = v.select(v.frame, sum(v.c2, order_by=base_t, group_by=v.pos)).show(100)

        schema = {
            'c2': IntType(nullable=False),
            'c3': FloatType(nullable=False),
            'c4': BoolType(nullable=False),
        }
        new_t = cl.create_table('insert_test', schema=schema)
        new_t.add_column(c2_sum=sum(new_t.c2, group_by=new_t.c4, order_by=new_t.c3))
        rows = list(t.select(t.c2, t.c4, t.c3).collect())
        new_t.insert(rows)
        _ = new_t.show(0)

    def test_aggregates(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        _ = t[t.c2 % 2, sum(t.c2), count(t.c2), sum(t.c2) + count(t.c2), sum(t.c2) + (t.c2 % 2)]\
            .group_by(t.c2 % 2).show()

        # check that aggregates don't show up in the wrong places
        with pytest.raises(excs.Error):
            # aggregate in where clause
            _ = t[sum(t.c2) > 0][sum(t.c2)].group_by(t.c2 % 2).show()
        with pytest.raises(excs.Error):
            # aggregate in group_by clause
            _ = t[sum(t.c2)].group_by(sum(t.c2)).show()
        with pytest.raises(excs.Error):
            # mixing aggregates and non-aggregates
            _ = t[sum(t.c2) + t.c2].group_by(t.c2 % 2).show()
        with pytest.raises(excs.Error):
            # nested aggregates
            _ = t[sum(count(t.c2))].group_by(t.c2 % 2).show()

    def test_udas(self, test_tbl: catalog.Table) -> None:
        t = test_tbl

        @pxt.uda(
            name='window_agg', init_types=[IntType()], update_types=[IntType()], value_type=IntType(),
            allows_window=True, requires_order_by=False)
        class WindowAgg:
            def __init__(self, val: int = 0):
                self.val = val
            def update(self, ignore: int) -> None:
                pass
            def value(self) -> int:
                return self.val

        @pxt.uda(
            name='ordered_agg', init_types=[IntType()], update_types=[IntType()], value_type=IntType(),
            requires_order_by=True, allows_window=True)
        class WindowAgg:
            def __init__(self, val: int = 0):
                self.val = val
            def update(self, i: int) -> None:
                pass
            def value(self) -> int:
                return self.val

        @pxt.uda(
            name='std_agg', init_types=[IntType()], update_types=[IntType()], value_type=IntType(),
            requires_order_by=False, allows_window=False)
        class StdAgg:
            def __init__(self, val: int = 0):
                self.val = val
            def update(self, i: int) -> None:
                pass
            def value(self) -> int:
                return self.val

        # init arg is passed along
        assert t.select(out=window_agg(t.c2, order_by=t.c2)).collect()[0]['out'] == 0
        assert t.select(out=window_agg(t.c2, val=1, order_by=t.c2)).collect()[0]['out'] == 1

        with pytest.raises(excs.Error) as exc_info:
            _ = t.select(window_agg(t.c2, val=t.c2, order_by=t.c2)).collect()
        assert 'needs to be a constant' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            # ordering expression not a pixeltable expr
            _ = t.select(ordered_agg(1, t.c2)).collect()
        assert 'but instead is a' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            # explicit order_by
            _ = t.select(ordered_agg(t.c2, order_by=t.c2)).collect()
        assert 'order_by invalid' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            # order_by for non-window function
            _ = t.select(std_agg(t.c2, order_by=t.c2)).collect()
        assert 'does not allow windows' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            # group_by for non-window function
            _ = t.select(std_agg(t.c2, group_by=t.c4)).collect()
        assert 'group_by invalid' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            # missing init type
            @pxt.uda(update_types=[IntType()], value_type=IntType())
            class WindowAgg:
                def __init__(self, val: int = 0):
                    self.val = val
                def update(self, ignore: int) -> None:
                    pass
                def value(self) -> int:
                    return self.val
        assert 'init_types must be a list of' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            # missing update parameter
            @pxt.uda(init_types=[IntType()], update_types=[], value_type=IntType())
            class WindowAgg:
                def __init__(self, val: int = 0):
                    self.val = val
                def update(self) -> None:
                    pass
                def value(self) -> int:
                    return self.val
        assert 'must have at least one parameter' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            # missing update type
            @pxt.uda(init_types=[IntType()], update_types=[IntType()], value_type=IntType())
            class WindowAgg:
                def __init__(self, val: int = 0):
                    self.val = val
                def update(self, i1: int, i2: int) -> None:
                    pass
                def value(self) -> int:
                    return self.val
        assert 'update_types must be a list of' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            # duplicate parameter names
            @pxt.uda(init_types=[IntType()], update_types=[IntType()], value_type=IntType())
            class WindowAgg:
                def __init__(self, val: int = 0):
                    self.val = val
                def update(self, val: int) -> None:
                    pass
                def value(self) -> int:
                    return self.val
        assert 'cannot have parameters with the same name: val' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            # invalid name
            @pxt.uda(name='not an identifier', init_types=[IntType()], update_types=[IntType()], value_type=IntType())
            class WindowAgg:
                def __init__(self, val: int = 0):
                    self.val = val
                def update(self, i1: int, i2: int) -> None:
                    pass
                def value(self) -> int:
                    return self.val
        assert 'invalid name' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            # reserved parameter name
            @pxt.uda(init_types=[IntType()], update_types=[IntType()], value_type=IntType())
            class WindowAgg:
                def __init__(self, val: int = 0):
                    self.val = val
                def update(self, order_by: int) -> None:
                    pass
                def value(self) -> int:
                    return self.val
        assert 'order_by is reserved' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            # reserved parameter name
            @pxt.uda(init_types=[IntType()], update_types=[IntType()], value_type=IntType())
            class WindowAgg:
                def __init__(self, val: int = 0):
                    self.val = val
                def update(self, group_by: int) -> None:
                    pass
                def value(self) -> int:
                    return self.val
        assert 'group_by is reserved' in str(exc_info.value).lower()



from typing import Dict, Any, List, Tuple

import PIL
import numpy as np
import pandas as pd
import pytest

import pixeltable as pxt
from pixeltable import exceptions as excs
from pixeltable.iterators import ComponentIterator
from pixeltable.iterators.video import FrameIterator
from pixeltable.tests.utils import assert_resultset_eq, get_video_files
from pixeltable.type_system import IntType, VideoType, JsonType


class ConstantImgIterator(ComponentIterator):
    """Component iterator that generates a fixed number of all-black 1280x720 images."""
    def __init__(self, video: str, num_frames: int = 10):
        self.img = PIL.Image.new('RGB', (1280, 720))
        self.next_frame_idx = 0
        self.num_frames = num_frames
        self.pos_msec = 0.0
        self.pos_frame = 0.0

    @classmethod
    def input_schema(cls) -> Dict[str, pxt.ColumnType]:
        return {
            'video': VideoType(nullable=False),
            'fps': pxt.FloatType()
        }

    @classmethod
    def output_schema(cls, *args: Any, **kwargs: Any) -> Tuple[Dict[str, pxt.ColumnType], List[str]]:
        return {
            'frame_idx': IntType(),
            'pos_msec': pxt.FloatType(),
            'pos_frame': pxt.FloatType(),
            'frame': pxt.ImageType(),
        }, ['frame']

    def __next__(self) -> Dict[str, Any]:
        while True:
            if self.next_frame_idx == self.num_frames:
                raise StopIteration
            result = {
                'frame_idx': self.next_frame_idx,
                'pos_msec': self.pos_msec,
                'pos_frame': self.pos_frame,
                'frame': self.img,
            }
            self.next_frame_idx += 1
            return result

    def close(self) -> None:
        pass

    def set_pos(self, pos: int) -> None:
        if pos == self.next_frame_idx:
            return
        self.next_frame_idx = pos


class TestComponentView:
    def test_basic(self, test_client: pxt.Client) -> None:
        cl = test_client
        # create video table
        schema = {'video': VideoType(), 'angle': IntType(), 'other_angle': IntType()}
        video_t = cl.create_table('video_tbl', schema)
        video_filepaths = get_video_files()

        # cannot add 'pos' column
        with pytest.raises(excs.Error) as excinfo:
            video_t.add_column(pos=IntType())
        assert 'reserved' in str(excinfo.value)

        # parameter missing
        with pytest.raises(excs.Error) as excinfo:
            args = {'fps': 1}
            _ = cl.create_view('test_view', video_t, iterator_class=FrameIterator, iterator_args=args)
        assert 'missing a required argument' in str(excinfo.value)

        # bad parameter type
        with pytest.raises(excs.Error) as excinfo:
            args = {'video': video_t.video, 'fps': '1'}
            _ = cl.create_view('test_view', video_t, iterator_class=FrameIterator, iterator_args=args)
        assert 'expected float' in str(excinfo.value)

        # bad parameter type
        with pytest.raises(excs.Error) as excinfo:
            args = {'video': 1, 'fps': 1}
            _ = cl.create_view('test_view', video_t, iterator_class=FrameIterator, iterator_args=args)
        assert 'expected file path' in str(excinfo.value)

        # create frame view
        args = {'video': video_t.video, 'fps': 1}
        view_t = cl.create_view('test_view', video_t, iterator_class=FrameIterator, iterator_args=args)
        # computed column that references a column from the base
        view_t.add_column(angle2=view_t.angle + 1)
        # computed column that references an unstored and a stored computed view column
        view_t.add_column(v1=view_t.frame.rotate(view_t.angle2), stored=True)
        # computed column that references a stored computed column from the view
        view_t.add_column(v2=view_t.frame_idx - 1)
        # computed column that references an unstored view column and a column from the base; the stored value
        # cannot be materialized in SQL directly
        view_t.add_column(v3=view_t.frame.rotate(video_t.other_angle), stored=True)

        # and load data
        rows = [{'video': p, 'angle': 30, 'other_angle': -30} for p in video_filepaths]
        status = video_t.insert(rows)
        assert status.num_excs == 0
        # pos and frame_idx are identical
        res = view_t.select(view_t.pos, view_t.frame_idx).collect().to_pandas()
        assert np.all(res['pos'] == res['frame_idx'])

        video_url = video_t.select(video_t.video.fileurl).show(0)[0, 0]
        result = view_t.where(view_t.video == video_url).select(view_t.frame, view_t.frame_idx) \
            .collect()
        result = view_t.where(view_t.video == video_url).select(view_t.frame_idx).order_by(view_t.frame_idx) \
            .collect().to_pandas()
        assert len(result) > 0
        assert np.all(result['frame_idx'] == pd.Series(range(len(result))))

    def test_add_column(self, test_client: pxt.Client) -> None:
        cl = test_client
        # create video table
        video_t = cl.create_table('video_tbl', {'video': VideoType()})
        video_filepaths = get_video_files()
        # create frame view
        args = {'video': video_t.video, 'fps': 1}
        view_t = cl.create_view('test_view', video_t, iterator_class=FrameIterator, iterator_args=args)

        rows = [{'video': p} for p in video_filepaths]
        video_t.insert(rows)
        # adding a non-computed column backfills it with nulls
        view_t.add_column(annotation=JsonType(nullable=True))
        assert view_t.count() == view_t.where(view_t.annotation == None).count()
        # adding more data via the base table sets the column values to null
        video_t.insert(rows)
        _ = view_t.where(view_t.annotation == None).count()
        assert view_t.count() == view_t.where(view_t.annotation == None).count()

        with pytest.raises(excs.Error) as excinfo:
            view_t.add_column(annotation=JsonType(nullable=False))
        assert 'must be nullable' in str(excinfo.value)

    def test_update(self, test_client: pxt.Client) -> None:
        cl = test_client
        # create video table
        video_t = cl.create_table('video_tbl', {'video': VideoType()})
        # create frame view with manually updated column
        args = {'video': video_t.video, 'fps': 1}
        view_t = cl.create_view(
            'test_view', video_t, schema={'annotation': JsonType(nullable=True)},
            iterator_class=FrameIterator, iterator_args=args)

        video_filepaths = get_video_files()
        rows = [{'video': p} for p in video_filepaths]
        status = video_t.insert(rows)
        assert status.num_excs == 0
        import urllib
        video_url = urllib.parse.urljoin('file:', urllib.request.pathname2url(video_filepaths[0]))
        status = view_t.update({'annotation': {'a': 1}}, where=view_t.video == video_url)
        c1 = view_t.where(view_t.annotation != None).count()
        c2 = view_t.where(view_t.video == video_url).count()
        assert c1 == c2

        with pytest.raises(excs.Error) as excinfo:
            _ = cl.create_view(
                'bad_view', video_t, schema={'annotation': JsonType(nullable=False)},
                iterator_class=FrameIterator, iterator_args=args)
        assert 'must be nullable' in str(excinfo.value)

    # break up the snapshot tests for better (future) parallelization
    def test_snapshot1(self, test_client: pxt.Client) -> None:
        has_column = False
        has_filter  = False
        for reload_md in [False, True]:
            cl = pxt.Client(reload=True)
            self.run_snapshot_test(cl, has_column=has_column, has_filter=has_filter, reload_md=reload_md)

    def test_snapshot2(self, test_client: pxt.Client) -> None:
        has_column = True
        has_filter  = False
        for reload_md in [False, True]:
            cl = pxt.Client(reload=True)
            self.run_snapshot_test(cl, has_column=has_column, has_filter=has_filter, reload_md=reload_md)

    def test_snapshot3(self, test_client: pxt.Client) -> None:
        has_column = False
        has_filter  = True
        for reload_md in [False, True]:
            cl = pxt.Client(reload=True)
            self.run_snapshot_test(cl, has_column=has_column, has_filter=has_filter, reload_md=reload_md)

    def test_snapshot4(self, test_client: pxt.Client) -> None:
        has_column = True
        has_filter  = True
        for reload_md in [False, True]:
            cl = pxt.Client(reload=True)
            self.run_snapshot_test(cl, has_column=has_column, has_filter=has_filter, reload_md=reload_md)

    def run_snapshot_test(self, cl: pxt.Client, has_column: bool, has_filter: bool, reload_md: bool) -> None:
        base_path = 'video_tbl'
        view_path = 'test_view'
        snap_path = 'test_snap'

        # create video table
        video_t = cl.create_table(base_path, {'video': VideoType(), 'margin': IntType()})
        video_filepaths = get_video_files()
        rows = [{'video': path, 'margin': i * 10} for i, path in enumerate(video_filepaths)]
        status = video_t.insert(rows)
        assert status.num_rows == len(rows)
        assert status.num_excs == 0

        # create frame view with a computed column
        args = {'video': video_t.video}
        view_t = cl.create_view(
            view_path, video_t, iterator_class=ConstantImgIterator, iterator_args=args, is_snapshot=False)
        view_t.add_column(
            cropped=view_t.frame.crop([view_t.margin, view_t.margin, view_t.frame.width, view_t.frame.height]),
            stored=True)
        snap_col_expr = [view_t.cropped.width * view_t.cropped.height] if has_column else []
        view_query = \
            view_t.select(
                    view_t.margin, view_t.frame.width, view_t.frame.height, view_t.cropped.width,
                    view_t.cropped.height, *snap_col_expr)\
                .order_by(view_t.video, view_t.pos)
        if has_filter:
            view_query = view_query.where(view_t.frame_idx < 10)
        orig_resultset = view_query.collect()

        # create snapshot of view
        filter = view_t.frame_idx < 10 if has_filter else None
        schema = {'c1': view_t.cropped.width * view_t.cropped.height} if has_column else {}
        snap_t = cl.create_view(snap_path, view_t, schema=schema, filter=filter, is_snapshot=True)
        snap_cols = [snap_t.c1] if has_column else []
        snap_query = \
            snap_t.select(
                    snap_t.margin, snap_t.frame.width, snap_t.frame.height, snap_t.cropped.width,
                    snap_t.cropped.height, *snap_cols)\
                .order_by(snap_t.video, snap_t.pos)
        assert_resultset_eq(snap_query.collect(), orig_resultset)

        if reload_md:
            cl = pxt.Client(reload=True)
            video_t = cl.get_table(base_path)
            snap_t = cl.get_table(snap_path)
            snap_cols = [snap_t.c1] if has_column else []
            snap_query = \
                snap_t.select(
                        snap_t.margin, snap_t.frame.width, snap_t.frame.height, snap_t.cropped.width,
                        snap_t.cropped.height, *snap_cols) \
                    .order_by(snap_t.video, snap_t.pos)

        # snapshot is unaffected by base insert()
        status = video_t.insert(rows)
        assert status.num_excs == 0
        assert_resultset_eq(snap_query.collect(), orig_resultset)

        # snapshot is unaffected by base update()
        status = video_t.update({'margin': video_t.margin + 1})
        assert status.num_excs == 0
        assert_resultset_eq(snap_query.collect(), orig_resultset)

        # snapshot is unaffected by base delete()
        status = video_t.delete()
        assert status.num_excs == 0
        assert_resultset_eq(snap_query.collect(), orig_resultset)

        cl.drop_table(snap_path)
        cl.drop_table(view_path)
        cl.drop_table(base_path)

    def test_chained_views(self, test_client: pxt.Client) -> None:
        """Component view followed by a standard view"""
        cl = test_client
        # create video table
        schema = {'video': VideoType(), 'int1': IntType(), 'int2': IntType()}
        video_t = cl.create_table('video_tbl', schema)
        video_filepaths = get_video_files()

        # create first view
        args = {'video': video_t.video}
        v1 = cl.create_view('test_view', video_t, iterator_class=ConstantImgIterator, iterator_args=args)
        # computed column that references stored base column
        v1.add_column(int3=v1.int1 + 1)
        # stored computed column that references an unstored and a stored computed view column
        v1.add_column(img1=v1.frame.crop([v1.int3, v1.int3, v1.frame.width, v1.frame.height]), stored=True)
        # computed column that references a stored computed view column
        v1.add_column(int4=v1.frame_idx + 1)
        # unstored computed column that references an unstored and a stored computed view column
        v1.add_column(img2=v1.frame.crop([v1.int4, v1.int4, v1.frame.width, v1.frame.height]), stored=False)

        # create second view
        v2 = cl.create_view('chained_view', v1)
        # computed column that references stored video_t column
        v2.add_column(int5=v2.int1 + 1)
        v2.add_column(int6=v2.int2 + 1)
        # stored computed column that references a stored base column and a stored computed view column;
        # indirectly references int1
        v2.add_column(img3=v2.img1.crop([v2.int5, v2.int5, v2.img1.width, v2.img1.height]), stored=True)
        # stored computed column that references an unstored base column and a manually updated column from video_t;
        # indirectly references int2
        v2.add_column(img4=v2.img2.crop([v2.int6, v2.int6, v2.img2.width, v2.img2.height]), stored=True)
        # comuted column that indirectly references int1 and int2
        v2.add_column(int7=v2.img3.width + v2.img4.width)

        def check_view():
            assert_resultset_eq(
                v1.select(v1.int3).order_by(v1.video, v1.pos).collect(),
                v1.select(v1.int1 + 1).order_by(v1.video, v1.pos).collect())
            assert_resultset_eq(
                v1.select(v1.int4).order_by(v1.video, v1.pos).collect(),
                v1.select(v1.frame_idx + 1).order_by(v1.video, v1.pos).collect())
            assert_resultset_eq(
                v1\
                    .select(v1.video, v1.img1.width, v1.img1.height)\
                    .order_by(v1.video, v1.pos).collect(),
                v1\
                    .select(v1.video, v1.frame.width - v1.int1 - 1, v1.frame.height - v1.int1 - 1)\
                    .order_by(v1.video, v1.pos).collect())
            assert_resultset_eq(
                v2.select(v2.int5).order_by(v2.video, v2.pos).collect(),
                v2.select(v2.int1 + 1).order_by(v2.video, v2.pos).collect())
            assert_resultset_eq(
                v2.select(v2.int6).order_by(v2.video, v2.pos).collect(),
                v2.select(v2.int2 + 1).order_by(v2.video, v2.pos).collect())
            assert_resultset_eq(
                v2 \
                    .select(v2.video, v2.img3.width, v2.img3.height) \
                    .order_by(v2.video, v2.pos).collect(),
                v2 \
                    .select(v2.video, v2.frame.width - v2.int1 * 2 - 2, v2.frame.height - v2.int1 * 2 - 2) \
                    .order_by(v2.video, v2.pos).collect())
            assert_resultset_eq(
                v2 \
                    .select(v2.video, v2.img4.width, v2.img4.height) \
                    .order_by(v2.video, v2.pos).collect(),
                v2 \
                    .select(
                        v2.video, v2.frame.width - v2.frame_idx - v2.int2 - 2,
                        v2.frame.height - v2.frame_idx - v2.int2 - 2) \
                    .order_by(v2.video, v2.pos).collect())
            assert_resultset_eq(
                v2.select(v2.int7).order_by(v2.video, v2.pos).collect(),
                v2.select(v2.img3.width + v2.img4.width).order_by(v2.video, v2.pos).collect())
            assert_resultset_eq(
                v2.select(v2.int7).order_by(v2.video, v2.pos).collect(),
                v2.select(v2.frame.width - v2.int1 * 2 - 2 + v2.frame.width - v2.frame_idx - v2.int2 - 2)\
                    .order_by(v2.video, v2.pos).collect())

        # load data
        rows = [{'video': p, 'int1': i, 'int2': len(video_filepaths) - i} for i, p in enumerate(video_filepaths)]
        status = video_t.insert(rows)
        assert status.num_rows == video_t.count() + v1.count() + v2.count()
        check_view()

        # update int1: propagates to int3, img1, int5, img3, int7
        # TODO: how to test that img4 doesn't get recomputed as part of the computation of int7?
        # need to collect more runtime stats (eg, called functions)
        import urllib
        video_url = urllib.parse.urljoin('file:', urllib.request.pathname2url(video_filepaths[0]))
        status = video_t.update({'int1': video_t.int1 + 1}, where=video_t.video == video_url)
        assert status.num_rows == 1 + v1.where(v1.video == video_url).count() + v2.where(v2.video == video_url).count()
        assert sorted('int1 int3 img1 int5 img3 int7'.split()) == sorted([str.split('.')[1] for str in status.updated_cols])
        check_view()

        # update int2: propagates to img4, int6, int7
        status = video_t.update({'int2': video_t.int2 + 1}, where=video_t.video == video_url)
        assert status.num_rows == 1 + v2.where(v2.video == video_url).count()
        assert sorted('int2 img4 int6 int7'.split()) == sorted([str.split('.')[1] for str in status.updated_cols])
        check_view()



from __future__ import annotations

import dataclasses
import logging
from uuid import UUID

import sqlalchemy as sql

from .schema_object import SchemaObject
from pixeltable.env import Env
from pixeltable.metadata import schema


_logger = logging.getLogger('pixeltable')

class Dir(SchemaObject):
    def __init__(self, id: UUID, parent_id: UUID, name: str):
        super().__init__(id, name, parent_id)

    @classmethod
    def display_name(cls) -> str:
        return 'directory'

    def move(self, new_name: str, new_dir_id: UUID) -> None:
        super().move(new_name, new_dir_id)
        with Env.get().engine.begin() as conn:
            dir_md = schema.DirMd(name=new_name)
            conn.execute(
                sql.update(schema.Dir.__table__)
                .values({schema.Dir.parent_id: self._dir_id, schema.Dir.md: dataclasses.asdict(dir_md)})
                .where(schema.Dir.id == self._id))




from typing import Any

from pixeltable.type_system import StringType
import pixeltable.func as func


@func.udf(return_type=StringType(), param_types=[StringType()])
def str_format(format_str: str, *args: Any, **kwargs: Any) -> str:
    """ Return a formatted version of format_str, using substitutions from args and kwargs:
    - {<int>} will be replaced by the corresponding element in args
    - {<key>} will be replaced by the corresponding value in kwargs
    """
    return format_str.format(*args, **kwargs)



from .aggregate_function import Aggregator, AggregateFunction, uda
from .batched_function import BatchedFunction, ExplicitBatchedFunction
from .callable_function import CallableFunction
from .expr_template_function import ExprTemplateFunction
from .function import Function
from .function_registry import FunctionRegistry
from .nos_function import NOSFunction
from .signature import Signature, Parameter, Batch
from .udf import udf, make_function, expr_udf



import datetime
import json
import logging
import os
import pathlib
import subprocess

import pgserver
import toml

import pixeltable as pxt
import pixeltable.metadata as metadata
from pixeltable.env import Env
from pixeltable.type_system import \
    StringType, IntType, FloatType, BoolType, TimestampType, JsonType

_logger = logging.getLogger('pixeltable')


class Dumper:

    def __init__(self, output_dir='target', db_name='pxtdump') -> None:
        self.output_dir = pathlib.Path(output_dir)
        shared_home = pathlib.Path(os.environ.get('PIXELTABLE_HOME', '~/.pixeltable')).expanduser()
        mock_home_dir = self.output_dir / '.pixeltable'
        mock_home_dir.mkdir(parents=True, exist_ok=True)
        os.environ['PIXELTABLE_HOME'] = str(mock_home_dir)
        os.environ['PIXELTABLE_CONFIG'] = str(shared_home / 'config.yaml')
        os.environ['PIXELTABLE_DB'] = db_name
        os.environ['PIXELTABLE_PGDATA'] = str(shared_home / 'pgdata')

        Env.get().set_up(reinit_db=True)
        self.cl = pxt.Client()
        self.cl.logging(level=logging.DEBUG, to_stdout=True)

    def dump_db(self) -> None:
        md_version = metadata.VERSION
        dump_file = self.output_dir / f'pixeltable-v{md_version:03d}-test.dump.gz'
        _logger.info(f'Creating database dump at: {dump_file}')
        pg_package_dir = os.path.dirname(pgserver.__file__)
        pg_dump_binary = f'{pg_package_dir}/pginstall/bin/pg_dump'
        _logger.info(f'Using pg_dump binary at: {pg_dump_binary}')
        with open(dump_file, 'wb') as dump:
            pg_dump_process = subprocess.Popen(
                [pg_dump_binary, Env.get().db_url, '-U', 'postgres', '-Fc'],
                stdout=subprocess.PIPE
            )
            subprocess.run(
                ["gzip", "-9"],
                stdin=pg_dump_process.stdout,
                stdout=dump,
                check=True
            )
        info_file = self.output_dir / f'pixeltable-v{md_version:03d}-test-info.toml'
        git_sha = subprocess.check_output(['git', 'rev-parse', 'HEAD']).decode('ascii').strip()
        user = os.environ.get('USER', os.environ.get('USERNAME'))
        info_dict = {'pixeltable-dump': {
            'metadata-version': md_version,
            'git-sha': git_sha,
            'datetime': datetime.datetime.utcnow(),
            'user': user
        }}
        with open(info_file, 'w') as info:
            toml.dump(info_dict, info)

    # TODO: Add additional features to the test DB dump (ideally it should exercise
    # every major pixeltable DB feature)
    def create_tables(self) -> None:
        schema = {
            'c1': StringType(nullable=False),
            'c1n': StringType(nullable=True),
            'c2': IntType(nullable=False),
            'c3': FloatType(nullable=False),
            'c4': BoolType(nullable=False),
            'c5': TimestampType(nullable=False),
            'c6': JsonType(nullable=False),
            'c7': JsonType(nullable=False),
        }
        t = self.cl.create_table('sample_table', schema, primary_key='c2')
        t.add_column(c8=[[1, 2, 3], [4, 5, 6]])

        # Add columns for .astype converters to ensure they're persisted properly
        t.add_column(c2_as_float=t.c2.astype(FloatType()))

        # Add columns for .apply
        t.add_column(c2_to_string=t.c2.apply(str))
        t.add_column(c6_to_string=t.c6.apply(json.dumps))
        t.add_column(c6_back_to_json=t.c6_to_string.apply(json.loads))

        num_rows = 100
        d1 = {
            'f1': 'test string 1',
            'f2': 1,
            'f3': 1.0,
            'f4': True,
            'f5': [1.0, 2.0, 3.0, 4.0],
            'f6': {
                'f7': 'test string 2',
                'f8': [1.0, 2.0, 3.0, 4.0],
            },
        }
        d2 = [d1, d1]

        c1_data = [f'test string {i}' for i in range(num_rows)]
        c2_data = [i for i in range(num_rows)]
        c3_data = [float(i) for i in range(num_rows)]
        c4_data = [bool(i % 2) for i in range(num_rows)]
        c5_data = [datetime.datetime.now()] * num_rows
        c6_data = []
        for i in range(num_rows):
            d = {
                'f1': f'test string {i}',
                'f2': i,
                'f3': float(i),
                'f4': bool(i % 2),
                'f5': [1.0, 2.0, 3.0, 4.0],
                'f6': {
                    'f7': 'test string 2',
                    'f8': [1.0, 2.0, 3.0, 4.0],
                },
            }
            c6_data.append(d)

        c7_data = [d2] * num_rows
        rows = [
            {
                'c1': c1_data[i],
                'c1n': c1_data[i] if i % 10 != 0 else None,
                'c2': c2_data[i],
                'c3': c3_data[i],
                'c4': c4_data[i],
                'c5': c5_data[i],
                'c6': c6_data[i],
                'c7': c7_data[i],
            }
            for i in range(num_rows)
        ]
        t.insert(rows)


def main() -> None:
    _logger.info("Creating pixeltable test artifact.")
    dumper = Dumper()
    dumper.create_tables()
    dumper.dump_db()


if __name__ == "__main__":
    main()



from pixeltable.catalog import is_valid_identifier, is_valid_path

class TestCatalog:
    """Tests for miscellanous catalog functions."""
    def test_valid_identifier(self) -> None:
        valid_ids = ['a', 'a1', 'a_1', 'a_']
        invalid_ids = ['', '_', '__', '_a', '1a', 'a.b', '.a', 'a-b']
        for valid_id in valid_ids:
            assert is_valid_identifier(valid_id), valid_ids

        for invalid_id in invalid_ids:
            assert not is_valid_identifier(invalid_id), invalid_ids

    def test_valid_path(self) -> None:
        assert is_valid_path('', empty_is_valid=True)
        assert not is_valid_path('', empty_is_valid=False)

        valid_paths = ['a', 'a_.b_', 'a.b.c', 'a.b.c.d']
        invalid_paths = ['.', '..', 'a.', '.a', 'a..b']
        
        for valid_path in valid_paths:
            assert is_valid_path(valid_path, empty_is_valid=False), valid_path
            assert is_valid_path(valid_path, empty_is_valid=True), valid_path

        for invalid_path in invalid_paths:
            assert not is_valid_path(invalid_path, empty_is_valid=False), invalid_path
            assert not is_valid_path(invalid_path, empty_is_valid=True), invalid_path


from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple
import enum

import sqlalchemy as sql

from .expr import Expr
from .column_ref import ColumnRef
from .row_builder import RowBuilder
from .data_row import DataRow
import pixeltable.catalog as catalog
import pixeltable.type_system as ts


class ColumnPropertyRef(Expr):
    """A reference to a property of a table column

    The properties themselves are type-specific and may or may not need to reference the underlying column data.
    """
    class Property(enum.Enum):
        ERRORTYPE = 0
        ERRORMSG = 1
        FILEURL = 2
        LOCALPATH = 3

    def __init__(self, col_ref: ColumnRef, prop: Property):
        super().__init__(ts.StringType(nullable=True))
        self.components = [col_ref]
        self.prop = prop
        self.id = self._create_id()

    def default_column_name(self) -> Optional[str]:
        return str(self).replace('.', '_')

    def _equals(self, other: ColumnRef) -> bool:
        return self.prop == other.prop

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('prop', self.prop.value)]

    @property
    def _col_ref(self) -> ColumnRef:
        return self.components[0]

    def __str__(self) -> str:
        return f'{self._col_ref}.{self.prop.name.lower()}'

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        if not self._col_ref.col.is_stored:
            return None
        if self.prop == self.Property.ERRORTYPE:
            assert self._col_ref.col.sa_errortype_col is not None
            return self._col_ref.col.sa_errortype_col
        if self.prop == self.Property.ERRORMSG:
            assert self._col_ref.col.sa_errormsg_col is not None
            return self._col_ref.col.sa_errormsg_col
        if self.prop == self.Property.FILEURL:
            # the file url is stored as the column value
            return self._col_ref.sql_expr()
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        assert self.prop == self.Property.FILEURL or self.prop == self.Property.LOCALPATH
        assert data_row.has_val[self._col_ref.slot_idx]
        if self.prop == self.Property.FILEURL:
            data_row[self.slot_idx] = data_row.file_urls[self._col_ref.slot_idx]
        if self.prop == self.Property.LOCALPATH:
            data_row[self.slot_idx] = data_row.file_paths[self._col_ref.slot_idx]

    def _as_dict(self) -> Dict:
        return {'prop': self.prop.value, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'prop' in d
        return cls(components[0], cls.Property(d['prop']))




from __future__ import annotations
from typing import Optional, List, Tuple, Dict
from collections import OrderedDict, defaultdict, namedtuple
import os
import glob
from pathlib import Path
from time import time
import logging
from uuid import UUID
import hashlib

from pixeltable.env import Env


_logger = logging.getLogger('pixeltable')

class CacheEntry:
    def __init__(self, key: str, tbl_id: UUID, col_id: int, size: int, last_accessed_ts: int, ext: str):
        self.key = key
        self.tbl_id = tbl_id
        self.col_id = col_id
        self.size = size
        self.last_accessed_ts = last_accessed_ts
        self.ext = ext

    def path(self) -> Path:
        return Env.get().file_cache_dir / f'{self.tbl_id.hex}_{self.col_id}_{self.key}{self.ext}'

    @classmethod
    def from_file(cls, path: Path) -> CacheEntry:
        components = path.stem.split('_')
        assert len(components) == 3
        tbl_id = UUID(components[0])
        col_id = int(components[1])
        key = components[2]
        file_info = os.stat(str(path))
        return cls(key, tbl_id, col_id, file_info.st_size, file_info.st_mtime, path.suffix)


class FileCache:
    """
    A local cache of external (eg, S3) file references in cells of a stored table (ie, table or view).

    Cache entries are identified by a hash of the file url and stored in Env.filecache_dir. The time of last
    access of a cache entries is its file's mtime.

    TODO:
    - enforce a maximum capacity with LRU eviction
    - implement MRU eviction for queries that exceed the capacity
    """
    _instance: Optional[FileCache] = None
    ColumnStats = namedtuple('FileCacheColumnStats', ['tbl_id', 'col_id', 'num_files', 'total_size'])
    CacheStats = namedtuple(
        'FileCacheStats', ['total_size', 'num_requests', 'num_hits', 'num_evictions', 'column_stats'])

    @classmethod
    def get(cls) -> FileCache:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self.cache: OrderedDict[str, CacheEntry] = OrderedDict()  # ordered by entry.last_accessed_ts
        self.total_size = 0
        #self.capacity = Env.get().max_filecache_size
        self.num_requests = 0
        self.num_hits = 0
        self.num_evictions = 0
        paths = glob.glob(str(Env.get().file_cache_dir / '*'))
        entries = [CacheEntry.from_file(Path(path_str)) for path_str in paths]
        # we need to insert entries in order of last_accessed_ts
        entries.sort(key=lambda e: e.last_accessed_ts)
        for entry in entries:
            self.cache[entry.key] = entry
            self.total_size += entry.size

    def avg_file_size(self) -> int:
        if len(self.cache) == 0:
            return 0
        return int(self.total_size / len(self.cache))

    def num_files(self, tbl_id: Optional[UUID] = None) -> int:
        if tbl_id is None:
            return len(self.cache)
        entries = [e for e in self.cache.values() if e.tbl_id == tbl_id]
        return len(entries)

    def clear(self, tbl_id: Optional[UUID] = None, capacity: Optional[int] = None) -> None:
        """
        For testing purposes: allow resetting capacity and stats.
        """
        self.num_requests, self.num_hits, self.num_evictions = 0, 0, 0
        entries = list(self.cache.values())  # list(): avoid dealing with values() return type
        if tbl_id is not None:
            entries = [e for e in entries if e.tbl_id == tbl_id]
            _logger.debug(f'clearing {len(entries)} entries from file cache for table {tbl_id}')
        else:
            _logger.debug(f'clearing {len(entries)} entries from file cache')
        for entry in entries:
            del self.cache[entry.key]
            self.total_size -= entry.size
            os.remove(entry.path())
        # if capacity is not None:
        #     self.capacity = capacity
        # else:
        #     # need to reset to default
        #     self.capacity = Env.get().max_filecache_size
        # _logger.debug(f'setting file cache capacity to {self.capacity}')

    def _url_hash(self, url: str) -> str:
        h = hashlib.sha256()
        h.update(url.encode())
        return h.hexdigest()

    def lookup(self, url: str) -> Optional[Path]:
        self.num_requests += 1
        key = self._url_hash(url)
        entry = self.cache.get(key, None)
        if entry is None:
            _logger.debug(f'file cache miss for {url}')
            return None
        # update mtime and cache
        path = entry.path()
        path.touch(exist_ok=True)
        file_info = os.stat(str(path))
        entry.last_accessed_ts = file_info.st_mtime
        self.cache.move_to_end(key, last=True)
        self.num_hits += 1
        _logger.debug(f'file cache hit for {url}')
        return path

    # def can_admit(self, query_ts: int) -> bool:
    #     if self.total_size + self.avg_file_size <= self.capacity:
    #         return True
    #     assert len(self.cache) > 0
    #     # check whether we can evict the current lru entry
    #     lru_entry = next(iter(self.cache.values()))
    #     if lru_entry.last_accessed_ts >= query_ts:
    #         # the current query brought this entry in: we're not going to evict it
    #         return False
    #     return True

    def add(self, tbl_id: UUID, col_id: int, url: str, path: Path) -> Path:
        """Adds url at 'path' to cache and returns its new path.
        'path' will not be accessible after this call. Retains the extension of 'path'.
        """
        file_info = os.stat(str(path))
        _ = time()
        #if self.total_size + file_info.st_size > self.capacity:
        if False:
            if len(self.cache) == 0:
                # nothing to evict
                return
            # evict entries until we're below the limit or until we run into entries the current query brought in
            while True:
                lru_entry = next(iter(self.cache.values()))
                if lru_entry.last_accessed_ts >= query_ts:
                    # the current query brought this entry in: switch to MRU and ignore this put()
                    _logger.debug('file cache switched to MRU')
                    return
                self.cache.popitem(last=False)
                self.total_size -= lru_entry.size
                self.num_evictions += 1
                os.remove(str(lru_entry.path()))
                _logger.debug(f'evicted entry for cell {lru_entry.cell_id} from file cache')
                if self.total_size + file_info.st_size <= self.capacity:
                    break

        key = self._url_hash(url)
        assert key not in self.cache
        entry = CacheEntry(key, tbl_id, col_id, file_info.st_size, file_info.st_mtime, path.suffix)
        self.cache[key] = entry
        self.total_size += entry.size
        new_path = entry.path()
        os.rename(str(path), str(new_path))
        _logger.debug(f'added entry for cell {url} to file cache')
        return new_path

    def stats(self) -> CacheStats:
        # collect column stats
        # (tbl_id, col_id) -> (num_files, total_size)
        d: Dict[Tuple[int, int], List[int]] = defaultdict(lambda: [0, 0])
        for entry in self.cache.values():
            t = d[(entry.tbl_id, entry.col_id)]
            t[0] += 1
            t[1] += entry.size
        col_stats = [
            self.ColumnStats(tbl_id, col_id, num_files, size) for (tbl_id, col_id), (num_files, size) in d.items()
        ]
        col_stats.sort(key=lambda e: e[3], reverse=True)
        return self.CacheStats(self.total_size, self.num_requests, self.num_hits, self.num_evictions, col_stats)

    def debug_print(self) -> None:
        for entry in self.cache.values():
            print(f'CacheEntry: tbl_id={entry.tbl_id}, col_id={entry.col_id}, size={entry.size}')



from typing import Optional

import av

import pixeltable as pxt
import pixeltable.env as env
from pixeltable.tests.utils import get_video_files, get_audio_files
from pixeltable.type_system import VideoType, AudioType
from pixeltable.utils.media_store import MediaStore


class TestAudio:
    def check_audio_params(self, path: str, format: Optional[str] = None, codec: Optional[str] = None):
        with av.open(path) as container:
            audio_stream = container.streams.audio[0]
            if format is not None:
                assert format == container.format.name
            if codec is not None:
                assert codec == audio_stream.codec_context.codec.name

    def test_basic(self, test_client: pxt.Client) -> None:
        audio_filepaths = get_audio_files()
        cl = test_client
        audio_t = cl.create_table('audio', {'audio_file': AudioType()})
        status = audio_t.insert({'audio_file': p} for p in audio_filepaths)
        assert status.num_rows == len(audio_filepaths)
        assert status.num_excs == 0
        paths = audio_t.select(output=audio_t.audio_file.localpath).collect()['output']
        assert set(paths) == set(audio_filepaths)

    def test_extract(self, test_client: pxt.Client) -> None:
        video_filepaths = get_video_files()
        cl = test_client
        video_t = cl.create_table('videos', {'video': VideoType()})
        from pixeltable.functions.video import extract_audio
        video_t.add_column(audio=extract_audio(video_t.video))

        # one of the 3 videos doesn't have audio
        status = video_t.insert({'video': p} for p in video_filepaths)
        assert status.num_rows == len(video_filepaths)
        assert status.num_excs == 0
        assert MediaStore.count(video_t.get_id()) == len(video_filepaths) - 1
        assert video_t.where(video_t.audio != None).count() == len(video_filepaths) - 1
        assert env.Env.get().num_tmp_files() == 0

        # make sure everything works with a fresh client
        cl = pxt.Client()
        video_t = cl.get_table('videos')
        assert video_t.where(video_t.audio != None).count() == len(video_filepaths) - 1

        # test generating different formats and codecs
        paths = video_t.select(output=extract_audio(video_t.video, format='wav', codec='pcm_s16le')).collect()['output']
        # media files that are created as a part of a query end up in the tmp dir
        assert env.Env.get().num_tmp_files() == video_t.where(video_t.audio != None).count()
        for path in [p for p in paths if p is not None]:
            self.check_audio_params(path, format='wav', codec='pcm_s16le')
        # higher resolution
        paths = video_t.select(output=extract_audio(video_t.video, format='wav', codec='pcm_s32le')).collect()['output']
        for path in [p for p in paths if p is not None]:
            self.check_audio_params(path, format='wav', codec='pcm_s32le')

        for format in ['mp3', 'flac']:
            paths = video_t.select(output=extract_audio(video_t.video, format=format)).collect()['output']
            for path in [p for p in paths if p is not None]:
                self.check_audio_params(path, format=format)



from __future__ import annotations

import inspect
import json
import sys
from typing import Optional, List, Any, Dict, Tuple

import sqlalchemy as sql

import pixeltable.catalog as catalog
import pixeltable.exceptions as excs
import pixeltable.func as func
import pixeltable.type_system as ts
from .data_row import DataRow
from .expr import Expr
from .inline_array import InlineArray
from .inline_dict import InlineDict
from .row_builder import RowBuilder
from .rowid_ref import RowidRef


class FunctionCall(Expr):
    def __init__(
            self, fn: func.Function, bound_args: Dict[str, Any], order_by_clause: Optional[List[Any]] = None,
            group_by_clause: Optional[List[Any]] = None, is_method_call: bool = False):
        if order_by_clause is None:
            order_by_clause = []
        if group_by_clause is None:
            group_by_clause = []
        signature = fn.signature
        super().__init__(signature.get_return_type(bound_args))
        self.fn = fn
        self.is_method_call = is_method_call
        self.check_args(signature, bound_args)

        self.agg_init_args: Dict[str, Any] = {}
        if self.is_agg_fn_call:
            # we separate out the init args for the aggregator
            self.agg_init_args = {
                arg_name: arg for arg_name, arg in bound_args.items() if arg_name in fn.init_param_names
            }
            bound_args = {arg_name: arg for arg_name, arg in bound_args.items() if arg_name not in fn.init_param_names}

        # construct components, args, kwargs
        self.components: List[Expr] = []

        # Tuple[int, Any]:
        # - for Exprs: (index into components, None)
        # - otherwise: (-1, val)
        self.args: List[Tuple[int, Any]] = []
        self.kwargs: Dict[str, Tuple[int, Any]] = {}

        # we record the types of non-variable parameters for runtime type checks
        self.arg_types: List[ts.ColumnType] = []
        self.kwarg_types: Dict[str, ts.ColumnType] = {}
        # the prefix of parameters that are bound can be passed by position
        for param in fn.py_signature.parameters.values():
            if param.name not in bound_args or param.kind == inspect.Parameter.KEYWORD_ONLY:
                break
            arg = bound_args[param.name]
            if isinstance(arg, Expr):
                self.args.append((len(self.components), None))
                self.components.append(arg.copy())
            else:
                self.args.append((-1, arg))
            if param.kind != inspect.Parameter.VAR_POSITIONAL and param.kind != inspect.Parameter.VAR_KEYWORD:
                self.arg_types.append(signature.parameters[param.name].col_type)

        # the remaining args are passed as keywords
        kw_param_names = set(bound_args.keys()) - set(list(fn.py_signature.parameters.keys())[:len(self.args)])
        for param_name in kw_param_names:
            arg = bound_args[param_name]
            if isinstance(arg, Expr):
                self.kwargs[param_name] = (len(self.components), None)
                self.components.append(arg.copy())
            else:
                self.kwargs[param_name] = (-1, arg)
            if fn.py_signature.parameters[param_name].kind != inspect.Parameter.VAR_KEYWORD:
                self.kwarg_types[param_name] = signature.parameters[param_name].col_type

        # window function state:
        # self.components[self.group_by_start_idx:self.group_by_stop_idx] contains group_by exprs
        self.group_by_start_idx, self.group_by_stop_idx = 0, 0
        if len(group_by_clause) > 0:
            if isinstance(group_by_clause[0], catalog.Table):
                group_by_exprs = self._create_rowid_refs(group_by_clause[0])
            else:
                assert isinstance(group_by_clause[0], Expr)
                group_by_exprs = group_by_clause
            # record grouping exprs in self.components, we need to evaluate them to get partition vals
            self.group_by_start_idx = len(self.components)
            self.group_by_stop_idx = len(self.components) + len(group_by_exprs)
            self.components.extend(group_by_exprs)

        if isinstance(self.fn, func.ExprTemplateFunction):
            # we instantiate the template to create an Expr that can be evaluated and record that as a component
            fn_expr = self.fn.instantiate(**bound_args)
            self.components.append(fn_expr)
            self.fn_expr_idx = len(self.components) - 1
        else:
            self.fn_expr_idx = sys.maxsize

        # we want to make sure that order_by_clause get assigned slot_idxs, even though we won't need to evaluate them
        # (that's done in SQL)
        if len(order_by_clause) > 0 and not isinstance(order_by_clause[0], Expr):
            raise excs.Error(
                f'order_by argument needs to be a Pixeltable expression, but instead is a {type(order_by_clause[0])}')
        # don't add components after this, everthing after order_by_start_idx is part of the order_by clause
        self.order_by_start_idx = len(self.components)
        self.components.extend(order_by_clause)

        self.constant_args = {param_name for param_name, arg in bound_args.items() if not isinstance(arg, Expr)}
        # execution state for aggregate functions
        self.aggregator: Optional[Any] = None
        self.current_partition_vals: Optional[List[Any]] = None

        self.id = self._create_id()

    def _create_rowid_refs(self, tbl: catalog.Table) -> List[Expr]:
        target = tbl.tbl_version_path.tbl_version
        return [RowidRef(target, i) for i in range(target.num_rowid_columns())]

    @classmethod
    def check_args(cls, signature: func.Signature, bound_args: Dict[str, Any]) -> None:
        """Checks that bound_args are compatible with signature.

        Convert literals to the correct type and update bound_args in place, if necessary.
        """
        for param_name, arg in bound_args.items():
            param = signature.parameters[param_name]
            if isinstance(arg, dict):
                try:
                    arg = InlineDict(arg)
                    bound_args[param_name] = arg
                except excs.Error:
                    # this didn't work, but it might be a literal
                    pass
            if isinstance(arg, list) or isinstance(arg, tuple):
                try:
                    # If the column type is JsonType, force the literal to be JSON
                    arg = InlineArray(arg, force_json=param.col_type is not None and param.col_type.is_json_type())
                    bound_args[param_name] = arg
                except excs.Error:
                    # this didn't work, but it might be a literal
                    pass

            if not isinstance(arg, Expr):
                # make sure that non-Expr args are json-serializable and are literals of the correct type
                try:
                    _ = json.dumps(arg)
                except TypeError:
                    raise excs.Error(f"Argument for parameter '{param_name}' is not json-serializable: {arg}")
                if arg is not None:
                    try:
                        param_type = param.col_type
                        bound_args[param_name] = param_type.create_literal(arg)
                    except TypeError as e:
                        msg = str(e)
                        raise excs.Error(f"Argument for parameter '{param_name}': {msg[0].lower() + msg[1:]}")
                continue

            # variable parameters don't get type-checked, but they both need to be json-typed
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                assert isinstance(arg, InlineArray)
                arg.col_type = ts.JsonType()
                continue
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                assert isinstance(arg, InlineDict)
                arg.col_type = ts.JsonType()
            continue

            if not param_type.is_supertype_of(arg.col_type):
                raise excs.Error(
                    f'Parameter {param_name}: argument type {arg.col_type} does not match parameter type '
                    f'{param_type}')

    def is_nos_call(self) -> bool:
        return isinstance(self.fn, func.NOSFunction)

    def _equals(self, other: FunctionCall) -> bool:
        if self.fn != other.fn:
            return False
        if len(self.args) != len(other.args):
            return False
        for i in range(len(self.args)):
            if self.args[i] != other.args[i]:
                return False
        if self.group_by_start_idx != other.group_by_start_idx:
            return False
        if self.group_by_stop_idx != other.group_by_stop_idx:
            return False
        if self.order_by_start_idx != other.order_by_start_idx:
            return False
        return True

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [
            ('fn', id(self.fn)),  # use the function pointer, not the fqn, which isn't set for lambdas
            ('args', self.args),
            ('kwargs', self.kwargs),
            ('group_by_start_idx', self.group_by_start_idx),
            ('group_by_stop_idx', self.group_by_stop_idx),
            ('order_by_start_idx', self.order_by_start_idx)
        ]

    def __str__(self) -> str:
        return self.display_str()

    def display_str(self, inline: bool = True) -> str:
        if self.is_method_call:
            return f'{self.components[0]}.{self.fn.name}({self._print_args(1, inline)})'
        else:
            fn_name = self.fn.display_name if self.fn.display_name != '' else 'anonymous_fn'
            return f'{fn_name}({self._print_args()})'

    def _print_args(self, start_idx: int = 0, inline: bool = True) -> str:
        arg_strs = [
            str(arg) if idx == -1 else str(self.components[idx]) for idx, arg in self.args[start_idx:]
        ]
        def print_arg(arg: Any) -> str:
            return f"'{arg}'" if isinstance(arg, str) else str(arg)
        arg_strs.extend([
            f'{param_name}={print_arg(arg) if idx == -1 else str(self.components[idx])}'
            for param_name, (idx, arg) in self.kwargs.items()
        ])
        if len(self.order_by) > 0:
            if self.fn.requires_order_by:
                arg_strs.insert(0, Expr.print_list(self.order_by))
            else:
                arg_strs.append(f'order_by={Expr.print_list(self.order_by)}')
        if len(self.group_by) > 0:
            arg_strs.append(f'group_by={Expr.print_list(self.group_by)}')
        # TODO: figure out the function name
        separator = ', ' if inline else ',\n    '
        return separator.join(arg_strs)

    def has_group_by(self) -> List[Expr]:
        return self.group_by_stop_idx != 0

    @property
    def group_by(self) -> List[Expr]:
        return self.components[self.group_by_start_idx:self.group_by_stop_idx]

    @property
    def order_by(self) -> List[Expr]:
        return self.components[self.order_by_start_idx:]

    @property
    def is_window_fn_call(self) -> bool:
        return isinstance(self.fn, func.AggregateFunction) and self.fn.allows_window and \
            (not self.fn.allows_std_agg \
             or self.has_group_by() \
             or (len(self.order_by) > 0 and not self.fn.requires_order_by))

    def get_window_sort_exprs(self) -> Tuple[List[Expr], List[Expr]]:
        return self.group_by, self.order_by

    @property
    def is_agg_fn_call(self) -> bool:
        return isinstance(self.fn, func.AggregateFunction)

    def get_agg_order_by(self) -> List[Expr]:
        assert self.is_agg_fn_call
        return self.order_by

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        # TODO: implement for standard aggregate functions
        return None

    def reset_agg(self) -> None:
        """
        Init agg state
        """
        assert self.is_agg_fn_call
        assert isinstance(self.fn, func.AggregateFunction)
        self.aggregator = self.fn.agg_cls(**self.agg_init_args)

    def update(self, data_row: DataRow) -> None:
        """
        Update agg state
        """
        assert self.is_agg_fn_call
        args, kwargs = self._make_args(data_row)
        self.aggregator.update(*args, **kwargs)

    def _make_args(self, data_row: DataRow) -> Tuple[List[Any], Dict[str, Any]]:
        """Return args and kwargs, constructed for data_row"""
        kwargs: Dict[str, Any] = {}
        for param_name, (component_idx, arg) in self.kwargs.items():
            val = arg if component_idx == -1 else data_row[self.components[component_idx].slot_idx]
            param = self.fn.signature.parameters[param_name]
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                # expand **kwargs parameter
                kwargs.update(val)
            else:
                assert param.kind != inspect.Parameter.VAR_POSITIONAL
                kwargs[param_name] = val

        args: List[Any] = []
        for param_idx, (component_idx, arg) in enumerate(self.args):
            val = arg if component_idx == -1 else data_row[self.components[component_idx].slot_idx]
            param = self.fn.signature.parameters_by_pos[param_idx]
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                # expand *args parameter
                assert isinstance(val, list)
                args.extend(val)
            elif param.kind == inspect.Parameter.VAR_KEYWORD:
                # expand **kwargs parameter
                assert isinstance(val, dict)
                kwargs.update(val)
            else:
                args.append(val)
        return args, kwargs

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        args, kwargs = self._make_args(data_row)
        signature = self.fn.signature
        if signature.parameters is not None:
            # check for nulls
            for i in range(len(self.arg_types)):
                if args[i] is None and not self.arg_types[i].nullable:
                    # we can't evaluate this function
                    data_row[self.slot_idx] = None
                    return
            for param_name, param_type in self.kwarg_types.items():
                if kwargs[param_name] is None and not param_type.nullable:
                    # we can't evaluate this function
                    data_row[self.slot_idx] = None
                    return

        if isinstance(self.fn, func.ExprTemplateFunction):
            # we need to evaluate the template
            # TODO: can we get rid of this extra copy?
            fn_expr = self.components[self.fn_expr_idx]
            data_row[self.slot_idx] = data_row[fn_expr.slot_idx]
        elif isinstance(self.fn, func.CallableFunction):
            data_row[self.slot_idx] = self.fn.py_fn(*args, **kwargs)
        elif self.is_window_fn_call:
            if self.has_group_by():
                if self.current_partition_vals is None:
                    self.current_partition_vals = [None] * len(self.group_by)
                partition_vals = [data_row[e.slot_idx] for e in self.group_by]
                if partition_vals != self.current_partition_vals:
                    # new partition
                    self.aggregator = self.fn.agg_cls(**self.agg_init_args)
                    self.current_partition_vals = partition_vals
            elif self.aggregator is None:
                self.aggregator = self.fn.agg_cls(**self.agg_init_args)
            self.aggregator.update(*args)
            data_row[self.slot_idx] = self.aggregator.value()
        else:
            assert self.is_agg_fn_call
            data_row[self.slot_idx] = self.aggregator.value()

    def _as_dict(self) -> Dict:
        result = {
            'fn': self.fn.as_dict(), 'args': self.args, 'kwargs': self.kwargs,
            'group_by_start_idx': self.group_by_start_idx, 'group_by_stop_idx': self.group_by_stop_idx,
            'order_by_start_idx': self.order_by_start_idx,
            **super()._as_dict()
        }
        return result

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'fn' in d
        assert 'args' in d
        assert 'kwargs' in d
        # reassemble bound args
        fn = func.Function.from_dict(d['fn'])
        param_names = list(fn.signature.parameters.keys())
        bound_args = {param_names[i]: arg if idx == -1 else components[idx] for i, (idx, arg) in enumerate(d['args'])}
        bound_args.update(
            {param_name: val if idx == -1 else components[idx] for param_name, (idx, val) in d['kwargs'].items()})
        group_by_exprs = components[d['group_by_start_idx']:d['group_by_stop_idx']]
        order_by_exprs = components[d['order_by_start_idx']:]
        fn_call = cls(
            func.Function.from_dict(d['fn']), bound_args, group_by_clause=group_by_exprs,
            order_by_clause=order_by_exprs)
        return fn_call



from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple

import sqlalchemy as sql

from .expr import Expr
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.catalog as catalog
import pixeltable.type_system as ts

class Literal(Expr):
    def __init__(self, val: Any, col_type: Optional[ts.ColumnType] = None):
        if col_type is not None:
            val = col_type.create_literal(val)
        else:
            # try to determine a type for val
            col_type = ts.ColumnType.infer_literal_type(val)
            if col_type is None:
                raise TypeError(f'Not a valid literal: {val}')
        super().__init__(col_type)
        self.val = val
        self.id = self._create_id()

    def default_column_name(self) -> Optional[str]:
        return 'Literal'

    def __str__(self) -> str:
        if self.col_type.is_string_type() or self.col_type.is_timestamp_type():
            return f"'{self.val}'"
        return str(self.val)

    def _equals(self, other: Literal) -> bool:
        return self.val == other.val

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('val', self.val)]

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        # we need to return something here so that we can generate a Where clause for predicates
        # that involve literals (like Where c > 0)
        return sql.sql.expression.literal(self.val)

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        # this will be called, even though sql_expr() does not return None
        data_row[self.slot_idx] = self.val

    def _as_dict(self) -> Dict:
        return {'val': self.val, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'val' in d
        return cls(d['val'])



from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple
import copy

import sqlalchemy as sql
import numpy as np

from .expr import Expr
from .data_row import DataRow
from .inline_dict import InlineDict
from .row_builder import RowBuilder
import pixeltable.catalog as catalog
import pixeltable.type_system as ts


class InlineArray(Expr):
    """
    Array 'literal' which can use Exprs as values.

    The literal can be cast as either a pixeltable `ArrayType` or `JsonType`. If `force_json`
    is `True`, it will always be cast as a `JsonType`. If `force_json` is `False`, it will be cast as an
    `ArrayType` if it is a homogenous array of scalars or arrays, or a `JsonType` otherwise.
    """
    def __init__(self, elements: Tuple, force_json: bool = False):
        # we need to call this in order to populate self.components
        super().__init__(ts.ArrayType((len(elements),), ts.IntType()))

        # elements contains
        # - for Expr elements: (index into components, None)
        # - for non-Expr elements: (-1, value)
        self.elements: List[Tuple[int, Any]] = []
        for el in elements:
            el = copy.deepcopy(el)
            if isinstance(el, list):
                # If col_type is an ArrayType, we'll require it to be a multidimensional array
                # of the specified underlying type
                el = InlineArray(tuple(el), force_json)
            if isinstance(el, dict):
                el = InlineDict(el)
            if isinstance(el, Expr):
                self.elements.append((len(self.components), None))
                self.components.append(el)
            else:
                self.elements.append((-1, el))

        inferred_element_type = ts.InvalidType()
        for idx, val in self.elements:
            if idx >= 0:
                inferred_element_type = ts.ColumnType.supertype(inferred_element_type, self.components[idx].col_type)
            else:
                inferred_element_type = ts.ColumnType.supertype(inferred_element_type, ts.ColumnType.infer_literal_type(val))
            if inferred_element_type is None:
                break

        if force_json or inferred_element_type is None:
            # JSON conversion is forced, or there is no common supertype
            # TODO: make sure this doesn't contain Images
            self.col_type = ts.JsonType()
        elif inferred_element_type.is_scalar_type():
            self.col_type = ts.ArrayType((len(self.elements),), inferred_element_type)
        elif inferred_element_type.is_array_type():
            assert isinstance(inferred_element_type, ts.ArrayType)
            self.col_type = ts.ArrayType(
                (len(self.elements), *inferred_element_type.shape), ts.ColumnType.make_type(inferred_element_type.dtype))
        else:
            self.col_type = ts.JsonType()

        self.id = self._create_id()

    def __str__(self) -> str:
        elem_strs = [str(val) if val is not None else str(self.components[idx]) for idx, val in self.elements]
        return f'[{", ".join(elem_strs)}]'

    def _equals(self, other: InlineDict) -> bool:
        return self.elements == other.elements

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('elements', self.elements)]

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        result = [None] * len(self.elements)
        for i, (child_idx, val) in enumerate(self.elements):
            if child_idx >= 0:
                result[i] = data_row[self.components[child_idx].slot_idx]
            else:
                result[i] = copy.deepcopy(val)
        if self.col_type.is_array_type():
            data_row[self.slot_idx] = np.array(result)
        else:
            data_row[self.slot_idx] = result

    def _as_dict(self) -> Dict:
        return {'elements': self.elements, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'elements' in d
        arg: List[Any] = []
        for idx, val in d['elements']:
            if idx >= 0:
                arg.append(components[idx])
            else:
                arg.append(val)
        return cls(tuple(arg))



import base64
import io
from typing import Optional

import PIL.Image
import numpy as np

import pixeltable as pxt
import pixeltable.type_system as ts
from pixeltable import env
from pixeltable.func import Batch


@pxt.udf
def chat_completions(
        messages: list,
        model: str,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        max_tokens: Optional[int] = None,
        n: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        response_format: Optional[dict] = None,
        seed: Optional[int] = None,
        top_p: Optional[float] = None,
        temperature: Optional[float] = None
) -> dict:
    from openai._types import NOT_GIVEN
    result = env.Env.get().openai_client.chat.completions.create(
        messages=messages,
        model=model,
        frequency_penalty=frequency_penalty if frequency_penalty is not None else NOT_GIVEN,
        logit_bias=logit_bias if logit_bias is not None else NOT_GIVEN,
        max_tokens=max_tokens if max_tokens is not None else NOT_GIVEN,
        n=n if n is not None else NOT_GIVEN,
        presence_penalty=presence_penalty if presence_penalty is not None else NOT_GIVEN,
        response_format=response_format if response_format is not None else NOT_GIVEN,
        seed=seed if seed is not None else NOT_GIVEN,
        top_p=top_p if top_p is not None else NOT_GIVEN,
        temperature=temperature if temperature is not None else NOT_GIVEN
    )
    return result.dict()


@pxt.udf
def vision(
        prompt: str,
        image: PIL.Image.Image,
        model: str = 'gpt-4-vision-preview'
) -> str:
    bytes_arr = io.BytesIO()
    image.save(bytes_arr, format='png')
    b64_bytes = base64.b64encode(bytes_arr.getvalue())
    b64_encoded_image = b64_bytes.decode('utf-8')
    messages = [
        {'role': 'user',
         'content': [
             {'type': 'text', 'text': prompt},
             {'type': 'image_url', 'image_url': {
                 'url': f'data:image/png;base64,{b64_encoded_image}'
             }}
         ]}
    ]
    result = env.Env.get().openai_client.chat.completions.create(
        messages=messages,
        model=model
    )
    return result.choices[0].message.content


@pxt.udf
def moderations(input: str, model: Optional[str] = None) -> dict:
    result = env.Env().get().openai_client.moderations.create(input=input, model=model)
    return result.dict()


@pxt.udf(batch_size=32, return_type=ts.ArrayType((None,), dtype=ts.FloatType()))
def embeddings(input: Batch[str], *, model: str) -> Batch[np.ndarray]:
    result = env.Env().get().openai_client.embeddings.create(
        input=input,
        model=model,
        encoding_format='float'
    )
    embeddings = [
        np.array(data.embedding, dtype=np.float64)
        for data in result.data
    ]
    return embeddings



import logging

import PIL
import pytest

import pixeltable as pxt
from pixeltable import catalog
from pixeltable import exceptions as excs
from pixeltable.tests.utils import create_test_tbl, assert_resultset_eq
from pixeltable.type_system import IntType, FloatType, ImageType

logger = logging.getLogger('pixeltable')

class TestView:
    """
    TODO:
    - test tree of views
    - test consecutive component views

    """
    def create_tbl(self, cl: pxt.Client) -> catalog.InsertableTable:
        """Create table with computed columns"""
        t = create_test_tbl(cl)
        t.add_column(d1=t.c3 - 1)
        # add column that can be updated
        t.add_column(c10=FloatType(nullable=True))
        t.update({'c10': t.c3})
        # computed column that depends on two columns: exercise duplicate elimination during query construction
        t.add_column(d2=t.c3 - t.c10)
        return t

    def test_basic(self, test_client: pxt.Client) -> None:
        cl = test_client
        t = self.create_tbl(cl)

        # create view with filter and computed columns
        schema = {
            'v1': t.c3 * 2.0,
            'v2': t.c6.f5,
        }
        v = cl.create_view('test_view', t, schema=schema, filter=t.c2 < 10)
        # TODO: test repr more thoroughly
        _ = v.__repr__()
        assert_resultset_eq(
            v.select(v.v1).order_by(v.c2).collect(),
            t.select(t.c3 * 2.0).where(t.c2 < 10).order_by(t.c2).collect())
        # view-only query; returns the same result
        assert_resultset_eq(
            v.select(v.v1).order_by(v.v1).collect(),
            t.select(t.c3 * 2.0).where(t.c2 < 10).order_by(t.c2).collect())
        # computed columns that don't reference the base table
        v.add_column(v3=v.v1 * 2.0)
        v.add_column(v4=v.v2[0])

        def check_view(t: pxt.Table, v: pxt.Table) -> None:
            assert v.count() == t.where(t.c2 < 10).count()
            assert_resultset_eq(
                v.select(v.v1).order_by(v.c2).collect(),
                t.select(t.c3 * 2.0).where(t.c2 < 10).order_by(t.c2).collect())
            assert_resultset_eq(
                v.select(v.v3).order_by(v.c2).collect(),
                t.select(t.c3 * 4.0).where(t.c2 < 10).order_by(t.c2).collect())
            assert_resultset_eq(
                v.select(v.v4).order_by(v.c2).collect(),
                t.select(t.c6.f5[0]).where(t.c2 < 10).order_by(t.c2).collect())
        check_view(t, v)

        # check view md after reload
        cl = pxt.Client(reload=True)
        t = cl.get_table('test_tbl')
        v = cl.get_table('test_view')
        check_view(t, v)

        view_query = v.select(v.v1).order_by(v.c2)
        base_query = t.select(t.c3 * 2.0).where(t.c2 < 10).order_by(t.c2)

        # insert data: of 20 new rows, only 10 are reflected in the view
        rows = list(t.select(t.c1, t.c1n, t.c2, t.c3, t.c4, t.c5, t.c6, t.c7, t.c10).where(t.c2 < 20).collect())
        status = t.insert(rows)
        assert status.num_rows == 30
        assert t.count() == 120
        check_view(t, v)

        # update data: cascade to view
        status = t.update({'c4': True, 'c3': t.c3 + 1.0, 'c10': t.c10 - 1.0}, where=t.c2 < 5, cascade=True)
        assert status.num_rows == 10 * 2  # *2: rows affected in both base table and view
        assert t.count() == 120
        check_view(t, v)

        # base table delete is reflected in view
        status = t.delete(where=t.c2 < 5)
        status.num_rows == 10 * 2  # *2: rows affected in both base table and view
        assert t.count() == 110
        check_view(t, v)

        # test delete view
        cl.drop_table('test_view')
        with pytest.raises(excs.Error) as exc_info:
            _ = cl.get_table('test_view')
        assert 'No such path:' in str(exc_info.value)
        cl = pxt.Client(reload=True)
        # still true after reload
        with pytest.raises(excs.Error) as exc_info:
            _ = cl.get_table('test_view')
        assert 'No such path:' in str(exc_info.value)

        t = cl.get_table('test_tbl')
        with pytest.raises(excs.Error) as exc_info:
            _ = cl.create_view('lambda_view', t, schema={'v1': lambda c3: c3 * 2.0})
        assert 'computed with a callable' in str(exc_info.value).lower()

    def test_parallel_views(self, test_client: pxt.Client) -> None:
        """Two views over the same base table, with non-overlapping filters"""
        cl = test_client
        t = self.create_tbl(cl)

        # create view with filter and computed columns
        v1 = cl.create_view('v1', t, schema={'v1': t.c3 * 2}, filter=t.c2 < 10)
        # create another view with a non-overlapping filter and computed columns
        v2 = cl.create_view('v2', t, schema={'v1': t.c3 * 3}, filter=(t.c2 < 20) & (t.c2 >= 10))

        # sanity checks
        v1_query = v1.select(v1.v1).order_by(v1.c2)
        v2_query = v2.select(v2.v1).order_by(v2.c2)
        b1_query = t.select(t.c3 * 2).where(t.c2 < 10).order_by(t.c2)
        b2_query = t.select(t.c3 * 3).where((t.c2 >= 10) & (t.c2 < 20)).order_by(t.c2)
        assert_resultset_eq(v1_query.collect(), b1_query.collect())
        assert_resultset_eq(v2_query.collect(), b2_query.collect())

        # insert data: of 20 new rows, only 10 show up in each view
        rows = list(t.select(t.c1, t.c1n, t.c2, t.c3, t.c4, t.c5, t.c6, t.c7, t.c10).where(t.c2 < 20).collect())
        status = t.insert(rows)
        assert status.num_rows == 40
        assert t.count() == 120
        assert v1.count() == 20
        assert v2.count() == 20
        assert_resultset_eq(v1_query.collect(), b1_query.collect())
        assert_resultset_eq(v2_query.collect(), b2_query.collect())

        # update data: cascade to views
        status = t.update(
            {'c4': True, 'c3': t.c3 + 1, 'c10': t.c10 - 1.0}, where=(t.c2 >= 5) & (t.c2 < 15), cascade=True)
        assert status.num_rows == 20 * 2  # *2: rows affected in both base table and view
        assert t.count() == 120
        assert v1.count() == 20
        assert v2.count() == 20
        assert_resultset_eq(v1_query.collect(), b1_query.collect())
        assert_resultset_eq(v2_query.collect(), b2_query.collect())


        # base table delete is reflected in view
        status = t.delete(where=(t.c2 >= 5) & (t.c2 < 15))
        status.num_rows == 10 * 2  # *2: rows affected in both base table and view
        assert t.count() == 100
        assert v1.count() == 10
        assert v2.count() == 10
        assert_resultset_eq(v1_query.collect(), b1_query.collect())
        assert_resultset_eq(v2_query.collect(), b2_query.collect())

    def test_chained_views(self, test_client: pxt.Client) -> None:
        """Two views, the second one is a view over the first one"""
        cl = test_client
        t = self.create_tbl(cl)

        # create view with filter and computed columns
        v1 = cl.create_view('v1', t, schema={'col1': t.c3 * 2}, filter=t.c2 < 10)
        # create a view on top of v1
        v2_schema = {
            'col2': t.c3 * 3,  # only base
            'col3': v1.col1 / 2,  # only v1
            'col4': t.c10 + v1.col1,  # both base and v1
        }
        v2 = cl.create_view('v2', v1, schema=v2_schema, filter=t.c2 < 5)

        def check_views():
            assert_resultset_eq(
                v1.select(v1.col1).order_by(v1.c2).collect(),
                t.select(t.c3 * 2).where(t.c2 < 10).order_by(t.c2).collect())
            assert_resultset_eq(
                v2.select(v2.col1).order_by(v2.c2).collect(),
                v1.select(v1.col1).where(v1.c2 < 5).order_by(v1.c2).collect())
            assert_resultset_eq(
                v2.select(v2.col2).order_by(v2.c2).collect(),
                t.select(t.c3 * 3).where(t.c2 < 5).order_by(t.c2).collect())
            assert_resultset_eq(
                v2.select(v2.col3).order_by(v2.c2).collect(),
                v1.select(v1.col1 / 2).where(v1.c2 < 5).order_by(v2.c2).collect())
            assert_resultset_eq(
                v2.select(v2.col4).order_by(v2.c2).collect(),
                v1.select(v1.c10 + v1.col1).where(v1.c2 < 5).order_by(v1.c2).collect())
                #t.select(t.c10 * 2).where(t.c2 < 5).order_by(t.c2).collect())
        check_views()

        # insert data: of 20 new rows; 10 show up in v1, 5 in v2
        base_version, v1_version, v2_version = t.version(), v1.version(), v2.version()
        rows = list(t.select(t.c1, t.c1n, t.c2, t.c3, t.c4, t.c5, t.c6, t.c7, t.c10).where(t.c2 < 20).collect())
        status = t.insert(rows)
        assert status.num_rows == 20 + 10 + 5
        assert t.count() == 120
        assert v1.count() == 20
        assert v2.count() == 10
        # all versions were incremented
        assert t.version() == base_version + 1
        assert v1.version() == v1_version + 1
        assert v2.version() == v2_version + 1
        check_views()

        # update data: cascade to both views
        base_version, v1_version, v2_version = t.version(), v1.version(), v2.version()
        status = t.update({'c4': True, 'c3': t.c3 + 1}, where=t.c2 < 15, cascade=True)
        assert status.num_rows == 30 + 20 + 10
        assert t.count() == 120
        # all versions were incremented
        assert t.version() == base_version + 1
        assert v1.version() == v1_version + 1
        assert v2.version() == v2_version + 1
        check_views()

        # update data: cascade only to v2
        base_version, v1_version, v2_version = t.version(), v1.version(), v2.version()
        status = t.update({'c10': t.c10 - 1.0}, where=t.c2 < 15, cascade=True)
        assert status.num_rows == 30 + 10
        assert t.count() == 120
        # v1 did not get updated
        assert t.version() == base_version + 1
        assert v1.version() == v1_version
        assert v2.version() == v2_version + 1
        check_views()

        # base table delete is reflected in both views
        base_version, v1_version, v2_version = t.version(), v1.version(), v2.version()
        status = t.delete(where=t.c2 == 0)
        status.num_rows == 1 + 1 + 1
        assert t.count() == 118
        assert v1.count() == 18
        assert v2.count() == 8
        # all versions were incremented
        assert t.version() == base_version + 1
        assert v1.version() == v1_version + 1
        assert v2.version() == v2_version + 1
        check_views()

        # base table delete is reflected only in v1
        base_version, v1_version, v2_version = t.version(), v1.version(), v2.version()
        status = t.delete(where=t.c2 == 5)
        status.num_rows == 1 + 1
        assert t.count() == 116
        assert v1.count() == 16
        assert v2.count() == 8
        # v2 was not updated
        assert t.version() == base_version + 1
        assert v1.version() == v1_version + 1
        assert v2.version() == v2_version
        check_views()

    def test_unstored_columns(self, test_client: pxt.Client) -> None:
        """Test chained views with unstored columns"""
        # create table with image column and two updateable int columns
        cl = test_client
        schema = {
            'img': ImageType(),
            'int1': IntType(nullable=False),
            'int2': IntType(nullable=False),
        }
        t = cl.create_table('test_tbl', schema)
        # populate table with images of a defined size
        width, height = 100, 100
        rows = [
            {
                'img': PIL.Image.new('RGB', (width, height), color=(0, 0, 0)).tobytes('jpeg', 'RGB'),
                'int1': i,
                'int2': i,
            }
            for i in range(100)
        ]
        t.insert(rows)

        # view with unstored column that depends on int1 and a manually updated column (int4)
        v1_schema = {
            'img2': {
                'value': t.img.crop([t.int1, t.int1, width, height]),
                'stored': False,
            },
            'int3': t.int1 * 2,
            'int4': IntType(nullable=True),  # TODO: add default
        }
        logger.debug('******************* CREATE V1')
        v1 = cl.create_view('v1', t, schema=v1_schema)
        v1.update({'int4': 1})
        _ = v1.select(v1.img2.width, v1.img2.height).collect()

        # view with stored column that depends on t and view1
        v2_schema = {
            'img3': {
                # use the actual width and height of the image (not 100, which will pad the image)
                'value': v1.img2.crop([t.int1 + t.int2, v1.int3 + v1.int4, v1.img2.width, v1.img2.height]),
                'stored': True,
              },
        }
        logger.debug('******************* CREATE V2')
        v2 = cl.create_view('v2', v1, schema=v2_schema, filter=v1.int1 < 10)

        def check_views() -> None:
            assert_resultset_eq(
                v1.select(v1.img2.width, v1.img2.height).order_by(v1.int1).collect(),
                t.select(t.img.width - t.int1, t.img.height - t.int1).order_by(t.int1).collect())
            assert_resultset_eq(
                v2.select(v2.img3.width, v2.img3.height).order_by(v2.int1).collect(),
                v1.select(v1.img2.width - v1.int1 - v1.int2, v1.img2.height - v1.int3 - v1.int4)\
                    .where(v1.int1 < 10).order_by(v1.int1).collect())
        check_views()

        logger.debug('******************* INSERT')
        status = t.insert(rows, fail_on_exception=False)
        v1.update({'int4': 1}, where=v1.int4 == None)
        logger.debug('******************* POST INSERT')
        check_views()

        # update int1:
        # - cascades to v1 and v2
        # - removes a row from v2 (only 9 rows in t now qualify)
        logger.debug('******************* UPDATE INT1')
        t.update({'int1': t.int1 + 1})
        logger.debug('******************* POST UPDATE INT1')
        check_views()

        # update int2:
        # - cascades only to v2
        # - but requires join against v1 to access int4
        # TODO: but requires join against v1 to access int3 and int4
        logger.debug('******************* UPDATE INT2')
        t.update({'int2': t.int2 + 1})
        logger.debug('******************* POST UPDATE INT2')
        check_views()

    def test_computed_cols(self, test_client: pxt.Client) -> None:
        cl = test_client
        t = self.create_tbl(cl)

        # create view with computed columns
        schema = {
            'v1': t.c3 * 2.0,
            'v2': t.c6.f5,
        }
        v = cl.create_view('test_view', t, schema=schema)
        assert_resultset_eq(
            v.select(v.v1).order_by(v.c2).show(0),
            t.select(t.c3 * 2.0).order_by(t.c2).show(0))
        # computed columns that don't reference the base table
        v.add_column(v3=v.v1 * 2.0)
        v.add_column(v4=v.v2[0])

        # use view md after reload
        cl = pxt.Client(reload=True)
        t = cl.get_table('test_tbl')
        v = cl.get_table('test_view')

        # insert data
        rows = list(t.select(t.c1, t.c1n, t.c2, t.c3, t.c4, t.c5, t.c6, t.c7, t.c10).collect())
        t.insert(rows)
        assert t.count() == 200
        assert_resultset_eq(
            v.select(v.v1).order_by(v.c2).show(0),
            t.select(t.c3 * 2.0).order_by(t.c2).show(0))

        # update data: cascade to view
        t.update({'c4': True, 'c3': t.c3 + 1.0, 'c10': t.c10 - 1.0}, where=t.c2 < 5, cascade=True)
        assert t.count() == 200
        assert_resultset_eq(
            v.select(v.v1).order_by(v.c2).show(0),
            t.select(t.c3 * 2.0).order_by(t.c2).show(0))

        # base table delete is reflected in view
        t.delete(where=t.c2 < 5)
        assert t.count() == 190
        assert_resultset_eq(
            v.select(v.v1).order_by(v.c2).show(0),
            t.select(t.c3 * 2.0).order_by(t.c2).show(0))

    def test_filter(self, test_client: pxt.Client) -> None:
        cl = test_client
        t = create_test_tbl(cl)

        # create view with filter
        v = cl.create_view('test_view', t, filter=t.c2 < 10)
        assert_resultset_eq(
            v.order_by(v.c2).show(0),
            t.where(t.c2 < 10).order_by(t.c2).show(0))

        # use view md after reload
        cl = pxt.Client(reload=True)
        t = cl.get_table('test_tbl')
        v = cl.get_table('test_view')

        # insert data: of 20 new rows, only 10 are reflected in the view
        rows = list(t.select(t.c1, t.c1n, t.c2, t.c3, t.c4, t.c5, t.c6, t.c7).where(t.c2 < 20).collect())
        t.insert(rows)
        assert t.count() == 120
        assert_resultset_eq(
            v.order_by(v.c2).show(0),
            t.where(t.c2 < 10).order_by(t.c2).show(0))

        # update data
        t.update({'c4': True, 'c3': t.c3 + 1.0}, where=t.c2 < 5, cascade=True)
        assert t.count() == 120
        assert_resultset_eq(
            v.order_by(v.c2).show(0),
            t.where(t.c2 < 10).order_by(t.c2).show(0))

        # base table delete is reflected in view
        t.delete(where=t.c2 < 5)
        assert t.count() == 110
        assert_resultset_eq(
            v.order_by(v.c2).show(0),
            t.where(t.c2 < 10).order_by(t.c2).show(0))

    def test_view_of_snapshot(self, test_client: pxt.Client) -> None:
        """Test view over a snapshot"""
        cl = test_client
        t = self.create_tbl(cl)
        snap = cl.create_view('test_snap', t, is_snapshot=True)

        # create view with filter and computed columns
        schema = {
            'v1': snap.c3 * 2.0,
            'v2': snap.c6.f5,
        }
        v = cl.create_view('test_view', snap, schema=schema, filter=snap.c2 < 10)

        def check_view(s: pxt.Table, v: pxt.Table) -> None:
            assert v.count() == s.where(s.c2 < 10).count()
            assert_resultset_eq(
                v.select(v.v1).order_by(v.c2).collect(),
                s.select(s.c3 * 2.0).where(s.c2 < 10).order_by(s.c2).collect())
            assert_resultset_eq(
                v.select(v.v2).order_by(v.c2).collect(),
                s.select(s.c6.f5).where(s.c2 < 10).order_by(s.c2).collect())

        check_view(snap, v)
        # computed columns that don't reference the base table
        v.add_column(v3=v.v1 * 2.0)
        v.add_column(v4=v.v2[0])
        assert v.count() == t.where(t.c2 < 10).count()

        # use view md after reload
        cl = pxt.Client(reload=True)
        t = cl.get_table('test_tbl')
        snap = cl.get_table('test_snap')
        v = cl.get_table('test_view')

        # insert data: no changes to view
        rows = list(t.select(t.c1, t.c1n, t.c2, t.c3, t.c4, t.c5, t.c6, t.c7, t.c10).where(t.c2 < 20).collect())
        t.insert(rows)
        assert t.count() == 120
        check_view(snap, v)

        # update data: no changes to view
        t.update({'c4': True, 'c3': t.c3 + 1.0, 'c10': t.c10 - 1.0}, where=t.c2 < 5, cascade=True)
        assert t.count() == 120
        check_view(snap, v)

        # base table delete: no change to view
        t.delete(where=t.c2 < 5)
        assert t.count() == 110
        check_view(snap, v)

    def test_snapshots(self, test_client: pxt.Client) -> None:
        """Test snapshot of a view of a snapshot"""
        cl = test_client
        t = self.create_tbl(cl)
        s = cl.create_view('test_snap', t, is_snapshot=True)
        assert s.select(s.c2).order_by(s.c2).collect()['c2'] == t.select(t.c2).order_by(t.c2).collect()['c2']

        with pytest.raises(excs.Error) as exc_info:
            v = cl.create_view('test_view', s, schema={'v1': t.c3 * 2.0})
        assert 'value expression cannot be computed in the context of the base test_snap' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            v = cl.create_view('test_view', s, filter=t.c2 < 10)
        assert 'filter cannot be computed in the context of the base test_snap' in str(exc_info.value).lower()

        # create view with filter and computed columns
        schema = {
            'v1': s.c3 * 2.0,
            'v2': s.c6.f5,
        }
        v = cl.create_view('test_view', s, schema=schema, filter=s.c2 < 10)
        orig_view_cols = v.column_names()
        view_s = cl.create_view('test_view_snap', v, is_snapshot=True)
        assert set(view_s.column_names()) == set(orig_view_cols)

        def check(s1: pxt.Table, v: pxt.Table, s2: pxt.Table) -> None:
            assert s1.where(s1.c2 < 10).count() == v.count()
            assert v.count() == s2.count()
            assert_resultset_eq(
                s1.select(s1.c3 * 2.0, s1.c6.f5).where(s1.c2 < 10).order_by(s1.c2).collect(),
                v.select(v.v1, v.v2).order_by(v.c2).collect())
            assert_resultset_eq(
                v.select(v.c3, v.c6, v.v1, v.v2).order_by(v.c2).collect(),
                s2.select(s2.c3, s2.c6, s2.v1, s2.v2).order_by(s2.c2).collect())
        check(s, v, view_s)

        # add more columns
        v.add_column(v3=v.v1 * 2.0)
        v.add_column(v4=v.v2[0])
        check(s, v, view_s)
        assert set(view_s.column_names()) == set(orig_view_cols)

        # check md after reload
        cl = pxt.Client(reload=True)
        t = cl.get_table('test_tbl')
        view_s = cl.get_table('test_view_snap')
        check(s, v, view_s)
        assert set(view_s.column_names()) == set(orig_view_cols)

        # insert data: no changes to snapshot
        rows = list(t.select(t.c1, t.c1n, t.c2, t.c3, t.c4, t.c5, t.c6, t.c7, t.c10).where(t.c2 < 20).collect())
        t.insert(rows)
        assert t.count() == 120
        check(s, v, view_s)

        # update data: no changes to snapshot
        t.update({'c4': True, 'c3': t.c3 + 1.0, 'c10': t.c10 - 1.0}, where=t.c2 < 5, cascade=True)
        assert t.count() == 120
        check(s, v, view_s)

        # base table delete: no changes to snapshot
        t.delete(where=t.c2 < 5)
        assert t.count() == 110
        check(s, v, view_s)



import io
import pyarrow as pa
import pyarrow.parquet as pq
import torch
import torch.utils.data
import math
from pathlib import Path
import PIL.Image
import json
from typing import Dict, Tuple, Generator, Any
import datetime

from pixeltable.type_system import ColumnType
from pixeltable.utils.parquet import get_part_metadata
import numpy as np

def _cumsum(lst):
    acc = [0]
    for x in lst:
        acc.append(acc[-1] + x)
    return acc

class PixeltablePytorchDataset(torch.utils.data.IterableDataset):
    """
    PyTorch dataset interface for pixeltable data.
    NB. This class must inherit from torch.utils.data.IterableDataset for it
    to work with torch.utils.data.DataLoader.
    """
    def __init__(
        self,
        path: Path,
        image_format: str,
    ):
        """
        Args:
            path: path to directory containing parquet files
            image_format: 'np' or 'pt'. 'np' is RGB uint8 array,
                        'pt' is result of torchvision.transforms.ToTensor()
        """
        super().__init__()

        self.path = path
        self.image_format = image_format
        assert image_format in ["np", "pt"]
        column_type_path = path / '.pixeltable.column_types.json'
        assert column_type_path.exists(), f"missing {column_type_path}"
        with column_type_path.open() as f:
            column_types = json.load(f)
        self.column_types = {k: ColumnType.from_dict(v) for k, v in column_types.items()}
        self.part_metadata = get_part_metadata(self.path)
        self._totals = _cumsum([x[1] for x in self.part_metadata])

    def _get_start_position(self, row_start: int) -> Tuple[int, int]:
        """
        Returns the starting parquet file and row within that file for a given 'global' row number.
        based on the individual sizes of each part
        """
        assert row_start >= self._totals[0]
        assert row_start < self._totals[-1]

        prev_acc = 0
        for i, acc in enumerate(self._totals[1:], start=1):
            if acc > row_start:
                return (i - 1, row_start - prev_acc)
            prev_acc = acc

        assert False, "unreachable"

    def _unmarshall(self, k: str, v: Any) -> Any:
        if self.column_types[k].is_image_type():
            assert isinstance(v, bytes)
            im = PIL.Image.open(io.BytesIO(v))
            arr = np.array(im) # will copy data to guarantee "WRITEABLE" flag assertion below.
            assert arr.flags["WRITEABLE"]

            if self.image_format == "np":
                return arr

            assert self.image_format == "pt"
            import torchvision  # pylint: disable = import-outside-toplevel

            # use arr instead of im in ToTensor() to guarantee array input
            # to torch.from_numpy is writable. Using im is a suspected cause of
            # https://github.com/pixeltable/pixeltable/issues/69
            return torchvision.transforms.ToTensor()(arr)
        elif self.column_types[k].is_json_type():
            assert isinstance(v, str)
            return json.loads(v)
        elif self.column_types[k].is_array_type():
            assert isinstance(v, np.ndarray)
            # WRITEABLE is required for torch collate function, or undefined behavior
            if not v.flags["WRITEABLE"]:
                vout = v.copy()
                assert vout.flags["WRITEABLE"]
                return vout
            else:
                return v
        elif self.column_types[k].is_timestamp_type():
            # pytorch default collation only supports numeric types
            assert isinstance(v, datetime.datetime)
            return v.timestamp()
        else:
            assert not isinstance(v, np.ndarray) # all array outputs should be handled above
            return v

    def __iter__(self) -> Generator[Dict[str, Any], None, None]:
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:  # single-process data loading, return the full iterator
            start_row = 0
            end_row = self._totals[-1]
        else:  # in a worker process
            num_workers = [
                math.floor(self._totals[-1] / float(worker_info.num_workers))
                for _ in range(worker_info.num_workers)
            ]
            assert self._totals[-1] - sum(num_workers) < worker_info.num_workers
            for i in range(self._totals[-1] - sum(num_workers)):
                num_workers[i] += 1

            assert sum(num_workers) == self._totals[-1]
            start_rows = _cumsum(num_workers)
            start_row = start_rows[worker_info.id]
            end_row = start_rows[worker_info.id + 1]

        if start_row == end_row:
            return iter([])  # type: ignore
        else:
            return self._iter_range(start_row, end_row)

    def _iter_range(self, start_row : int, end_row : int) -> Generator[Dict[str, Any], None, None]:
        (part_no, iter_start) = self._get_start_position(start_row)
        total = end_row - start_row

        acc = 0
        part_pos = part_no
        iter_pos = iter_start

        def _to_column_dict(tab : pa.Table) -> Dict[str, Any]:
            column_dict = {}
            for k in tab.column_names:
                if self.column_types[k].is_array_type():
                    # treat array columns as numpy arrays to easily preserve numpy type
                    column_dict[k] = tab.column(k).to_numpy()
                else:
                    # for the rest, use pydict to preserve python types
                    column_dict[k] = tab.column(k).to_pylist()
            return column_dict

        tab: pa.Table = pq.read_table(self.path / self.part_metadata[part_no][0])
        column_dict = _to_column_dict(tab)
        assert tab.num_rows == self.part_metadata[part_no][1]

        while True:
            while iter_pos < tab.num_rows and acc < total:
                next_tup = {}
                for col_name, col_vals in column_dict.items():
                    raw_val = col_vals[iter_pos]
                    next_tup[col_name] = self._unmarshall(col_name, raw_val)

                yield next_tup
                acc += 1
                iter_pos += 1

            if acc == total:
                break

            # move on to next part
            part_pos += 1
            assert part_pos < len(self.part_metadata)
            iter_pos = 0
            tab = pq.read_table(self.path / self.part_metadata[part_pos][0])
            column_dict = _to_column_dict(tab)


import json
from typing import Optional, Dict, List, Tuple, Any

import sqlalchemy as sql

import pixeltable.type_system as ts
from .expr import DataRow, Expr
from .row_builder import RowBuilder


class TypeCast(Expr):
    """
    An `Expr` that represents a type conversion from an underlying `Expr` to
    a specified `ColumnType`.
    """
    def __init__(self, underlying: Expr, new_type: ts.ColumnType):
        super().__init__(new_type)
        self.components: List[Expr] = [underlying]
        self.id: Optional[int] = self._create_id()

    @property
    def _underlying(self):
        return self.components[0]

    def _equals(self, other: 'TypeCast') -> bool:
        # `TypeCast` has no properties beyond those captured by `Expr`.
        return True

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('new_type', self.col_type)]

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        """
        `sql_expr` is unimplemented for now, in order to sidestep potentially thorny
        questions about consistency of doing type conversions in both Python and Postgres.
        """
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        original_val = data_row[self._underlying.slot_idx]
        data_row[self.slot_idx] = self.col_type.create_literal(original_val)

    def _as_dict(self) -> Dict:
        return {'new_type': self.col_type.as_dict(), **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'new_type' in d
        assert len(components) == 1
        return cls(components[0], ts.ColumnType.from_dict(d['new_type']))

    def __str__(self) -> str:
        return f'{self._underlying}.astype({self.col_type})'



from __future__ import annotations
from typing import Optional, Dict, Iterable, Iterator

from .expr import Expr


class ExprSet:
    """A set that also supports indexed lookup (by slot_idx and Expr.id)"""
    def __init__(self, elements: Optional[Iterable[Expr]] = None):
        self.exprs: Dict[int, Expr] = {}  # Expr.id -> Expr
        if elements is not None:
            for e in elements:
                self.append(e)

    def append(self, expr: Expr) -> None:
        if expr.id in self.exprs:
            return
        self.exprs[expr.id] = expr

    def extend(self, elements: Iterable[Expr]) -> None:
        for e in elements:
            self.append(e)

    def __contains__(self, item: Expr) -> bool:
        return item.id in self.exprs

    def __len__(self) -> int:
        return len(self.exprs)

    def __iter__(self) -> Iterator[Expr]:
        return iter(self.exprs.values())

    def __getitem__(self, index: object) -> Optional[Expr]:
        assert isinstance(index, int) or isinstance(index, Expr)
        if isinstance(index, int):
            # return expr with matching slot_idx
            return list(self.exprs.values())[index]
        else:
            return self.exprs.get(index.id)



import pathlib
import tempfile

from pixeltable.utils.transactional_directory import transactional_directory


class MyException(Exception):
    pass

class TestTransactionalDirectory:
    def test_success(self) -> None:
        test_dir = pathlib.Path(tempfile.mkdtemp())
        assert test_dir.exists()
        final = test_dir / "test_success"
        assert not final.exists()
        with transactional_directory(final) as folder:
            assert folder.exists()
            (folder / "subfolder1").mkdir()
            with (folder / "test.txt").open("w") as f:
                f.write("test")

        assert final.exists()
        assert (final / "subfolder1").is_dir()
        assert (final / "test.txt").read_text() == "test"

    def test_failure(self) -> None:
        test_dir = pathlib.Path(tempfile.mkdtemp())
        assert test_dir.exists()
        final = test_dir / "test_failure"
        assert not final.exists()

        try:
            with transactional_directory(final) as folder:
                assert folder.exists()
                (folder / "subfolder1").mkdir()
                with (folder / "test.txt").open("w") as f:
                    f.write("test")
                raise MyException()
        except MyException:
            pass

        assert not final.exists()



from __future__ import annotations

import logging
from typing import Optional, List, Union
from uuid import UUID

import pixeltable
from .column import Column
from .globals import POS_COLUMN_NAME
from .table_version import TableVersion

_logger = logging.getLogger('pixeltable')

class TableVersionPath:
    """
    A TableVersionPath represents the sequence of TableVersions from a base table to a particular view:
    - for a base table: only includes that TableVersion
    - for a view: includes the TableVersion for the view and all its bases
    - multiple snapshots can reference the same TableVersion, but with different bases, which means that the
      graph of TableVersions is a DAG, not a tree (which is why we cannot embed the DAG into TableVersion directly)

    TableVersionPath contains all metadata needed to execute queries and updates against a particular version of a
    table/view.
    """

    def __init__(self, tbl_version: TableVersion, base: Optional[TableVersionPath] = None):
        assert tbl_version is not None
        self.tbl_version = tbl_version
        self.base = base

    def tbl_id(self) -> UUID:
        """Return the id of the table/view that this path represents"""
        return self.tbl_version.id

    def version(self) -> int:
        """Return the version of the table/view that this path represents"""
        return self.tbl_version.version

    def tbl_name(self) -> str:
        """Return the name of the table/view that this path represents"""
        return self.tbl_version.name

    def path_len(self) -> int:
        """Return the length of the path"""
        return 1 if self.base is None else 1 + self.base.path_len()

    def is_snapshot(self) -> bool:
        """Return True if this is a path of snapshot versions"""
        if not self.tbl_version.is_snapshot:
            return False
        return self.base.is_snapshot() if self.base is not None else True

    def is_view(self) -> bool:
        return self.tbl_version.is_view()

    def is_component_view(self) -> bool:
        return self.tbl_version.is_component_view()

    def is_insertable(self) -> bool:
        return self.tbl_version.is_insertable()

    def get_tbl_versions(self) -> List[TableVersion]:
        """Return all tbl versions"""
        if self.base is None:
            return [self.tbl_version]
        return [self.tbl_version] + self.base.get_tbl_versions()

    def get_bases(self) -> List[TableVersion]:
        """Return all tbl versions"""
        if self.base is None:
            return []
        return self.base.get_tbl_versions()

    def find_tbl_version(self, id: UUID) -> Optional[TableVersion]:
        """Return the matching TableVersion in the chain of TableVersions, starting with this one"""
        if self.tbl_version.id == id:
            return self.tbl_version
        if self.base is None:
            return None
        return self.base.find_tbl_version(id)

    def __getattr__(self, col_name: str) -> 'pixeltable.exprs.ColumnRef':
        """Return a ColumnRef for the given column name."""
        from pixeltable.exprs import ColumnRef, RowidRef
        if col_name == POS_COLUMN_NAME and self.is_component_view():
            return RowidRef(self.tbl_version, self.tbl_version.store_tbl.pos_col_idx)
        if col_name not in self.tbl_version.cols_by_name:
            if self.base is None:
                raise AttributeError(f'Column {col_name} unknown')
            return getattr(self.base, col_name)
        col = self.tbl_version.cols_by_name[col_name]
        return ColumnRef(col)

    def __getitem__(self, index: object) -> Union['pixeltable.exprs.ColumnRef', 'pixeltable.dataframe.DataFrame']:
        """Return a ColumnRef for the given column name, or a DataFrame for the given slice.
        """
        if isinstance(index, str):
            # basically <tbl>.<colname>
            return self.__getattr__(index)
        from pixeltable.dataframe import DataFrame
        return DataFrame(self).__getitem__(index)

    def columns(self) -> List[Column]:
        """Return all columns visible in this tbl version path, including columns from bases"""
        result = self.tbl_version.cols.copy()
        if self.base is not None:
            base_cols = self.base.columns()
            # we only include base columns that don't conflict with one of our column names
            result.extend([c for c in base_cols if c.name not in self.tbl_version.cols_by_name])
        return result

    def get_column(self, name: str, include_bases: bool = True) -> Optional[Column]:
        """Return the column with the given name, or None if not found"""
        col = self.tbl_version.cols_by_name.get(name)
        if col is not None:
            return col
        elif self.base is not None and include_bases:
            return self.base.get_column(name)
        else:
            return None

    def has_column(self, col: Column, include_bases: bool = True) -> bool:
        """Return True if this table has the given column.
        """
        assert col.tbl is not None
        if col.tbl.id == self.tbl_version.id and col.tbl.effective_version == self.tbl_version.effective_version \
                and col.id in self.tbl_version.cols_by_id:
            # the column is visible in this table version
            return True
        elif self.base is not None and include_bases:
            return self.base.has_column(col)
        else:
            return False



from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple, Set, Iterable
from dataclasses import dataclass
import time
import sys

from .expr import Expr
from .expr_set import ExprSet
from .data_row import DataRow
import pixeltable.utils as utils
import pixeltable.func as func
import pixeltable.exceptions as excs
import pixeltable.catalog as catalog


class ExecProfile:
    def __init__(self, row_builder: RowBuilder):
        self.eval_time = [0.0] * row_builder.num_materialized
        self.eval_count = [0] * row_builder.num_materialized
        self.row_builder = row_builder

    def print(self, num_rows: int) -> str:
        for i in range(self.row_builder.num_materialized):
            if self.eval_count[i] == 0:
                continue
            per_call_time = self.eval_time[i] / self.eval_count[i]
            calls_per_row = self.eval_count[i] / num_rows
            multiple_str = f'({calls_per_row}x)' if calls_per_row > 1 else ''
            print(f'{self.row_builder.unique_exprs[i]}: {utils.print_perf_counter_delta(per_call_time)} {multiple_str}')


@dataclass
class ColumnSlotIdx:
    """Info for how to locate materialized column in DataRow
    TODO: can this be integrated into RowBuilder directly?
    """
    col: catalog.Column
    slot_idx: int


class RowBuilder:
    """Create and populate DataRows and table rows from exprs and computed columns

    For ColumnRefs to unstored iterator columns:
    - in order for them to be executable, we also record the iterator args and pass them to the ColumnRef
    """

    @dataclass
    class EvalCtx:
        """Context for evaluating a set of target exprs"""
        slot_idxs: List[int]  # slot idxs of exprs needed to evaluate target exprs; does not contain duplicates
        exprs: List[Expr]  # exprs corresponding to slot_idxs
        target_slot_idxs: List[int]  # slot idxs of target exprs; might contain duplicates
        target_exprs: List[Expr]  # exprs corresponding to target_slot_idxs

    def __init__(
            self, output_exprs: List[Expr], columns: List[catalog.Column],
            indices: List[Tuple[catalog.Column, func.Function]], input_exprs: List[Expr]
    ):
        """
        Args:
            output_exprs: list of Exprs to be evaluated
            columns: list of columns to be materialized
            indices: list of embeddings to be materialized (Tuple[indexed column, embedding function])
        """
        self.unique_exprs = ExprSet()  # dependencies precede their dependents
        self.next_slot_idx = 0

        # record input and output exprs; make copies to avoid reusing execution state
        unique_input_exprs = [self._record_unique_expr(e.copy(), recursive=False) for e in input_exprs]
        self.input_expr_slot_idxs = {e.slot_idx for e in unique_input_exprs}

        # output exprs: all exprs the caller wants to materialize
        # - explicitly requested output_exprs
        # - values for computed columns
        # - embedding values for indices
        resolve_cols = set(columns)
        self.output_exprs = [
            self._record_unique_expr(e.copy().resolve_computed_cols(resolve_cols=resolve_cols), recursive=True)
            for e in output_exprs
        ]

        # record columns for create_table_row()
        from .column_ref import ColumnRef
        self.table_columns: List[ColumnSlotIdx] = []
        for col in columns:
            if col.is_computed:
                assert col.value_expr is not None
                # create a copy here so we don't reuse execution state and resolve references to computed columns
                expr = col.value_expr.copy().resolve_computed_cols(resolve_cols=resolve_cols)
                expr = self._record_unique_expr(expr, recursive=True)
                self.add_table_column(col, expr.slot_idx)
                self.output_exprs.append(expr)
            else:
                # record a ColumnRef so that references to this column resolve to the same slot idx
                ref = ColumnRef(col)
                ref = self._record_unique_expr(ref, recursive=False)
                self.add_table_column(col, ref.slot_idx)

        # record indices; indexed by slot_idx
        self.index_columns: List[catalog.Column] = []
        for col, embedding_fn in indices:
            # we assume that the parameter of the embedding function is a ref to an image column
            assert col.col_type.is_image_type()
            # construct expr to compute embedding; explicitly resize images to the required size
            target_img_type = next(iter(embedding_fn.signature.parameters.values())).col_type
            expr = embedding_fn(ColumnRef(col).resize(target_img_type.size))
            expr = self._record_unique_expr(expr, recursive=True)
            self.output_exprs.append(expr)
            if len(self.index_columns) <= expr.slot_idx:
                # pad to slot_idx
                self.index_columns.extend([None] * (expr.slot_idx - len(self.index_columns) + 1))
            self.index_columns[expr.slot_idx] = col

        # default eval ctx: all output exprs
        self.default_eval_ctx = self.create_eval_ctx(self.output_exprs, exclude=unique_input_exprs)

        # references to unstored iterator columns:
        # - those ColumnRefs need to instantiate iterators
        # - we create and record the iterator args here and pass them to their respective ColumnRefs
        # - we do this instead of simply recording the iterator args as a component of those ColumnRefs,
        #   because that would cause them to be evaluated for every single row
        # - the separate eval ctx allows the ColumnRef to materialize the iterator args only when the underlying
        #   iterated object changes
        col_refs = [e for e in self.unique_exprs if isinstance(e, ColumnRef)]
        def refs_unstored_iter_col(col_ref: ColumnRef) -> bool:
            tbl = col_ref.col.tbl
            return tbl.is_component_view() and tbl.is_iterator_column(col_ref.col) and not col_ref.col.is_stored
        unstored_iter_col_refs = [col_ref for col_ref in col_refs if refs_unstored_iter_col(col_ref)]
        component_views = [col_ref.col.tbl for col_ref in unstored_iter_col_refs]
        unstored_iter_args = {view.id: view.iterator_args.copy() for view in component_views}
        self.unstored_iter_args = \
            {id: self._record_unique_expr(arg, recursive=True) for id, arg in unstored_iter_args.items()}

        for col_ref in unstored_iter_col_refs:
            iter_arg_ctx = self.create_eval_ctx([unstored_iter_args[col_ref.col.tbl.id]])
            col_ref.set_iter_arg_ctx(iter_arg_ctx)

        # we guarantee that we can compute the expr DAG in a single front-to-back pass
        for i, expr in enumerate(self.unique_exprs):
            assert expr.slot_idx == i

        # record transitive dependencies (list of set of slot_idxs, indexed by slot_idx)
        self.dependencies: List[Set[int]] = [set() for _ in range(self.num_materialized)]
        for expr in self.unique_exprs:
            if expr.slot_idx in self.input_expr_slot_idxs:
                # this is input and therefore doesn't depend on other exprs
                continue
            for d in expr.dependencies():
                self.dependencies[expr.slot_idx].add(d.slot_idx)
                self.dependencies[expr.slot_idx].update(self.dependencies[d.slot_idx])

        # derive transitive dependents
        self.dependents: List[Set[int]] = [set() for _ in range(self.num_materialized)]
        for expr in self.unique_exprs:
            for d in self.dependencies[expr.slot_idx]:
                self.dependents[d].add(expr.slot_idx)

        # records the output_expr that a subexpr belongs to
        # (a subexpr can be shared across multiple output exprs)
        self.output_expr_ids: List[Set[int]] = [set() for _ in range(self.num_materialized)]
        for e in self.output_exprs:
            self._record_output_expr_id(e, e.slot_idx)

    def add_table_column(self, col: catalog.Column, slot_idx: int) -> None:
        """Record a column that is part of the table row"""
        self.table_columns.append(ColumnSlotIdx(col, slot_idx))

    def output_slot_idxs(self) -> List[ColumnSlotIdx]:
        """Return ColumnSlotIdx for output columns"""
        return self.table_columns

    def index_slot_idxs(self) -> List[ColumnSlotIdx]:
        """Return ColumnSlotIdx for index columns"""
        return [
            ColumnSlotIdx(self.output_columns[i], i) for i in range(len(self.index_columns))
            if self.output_columns[i] is not None
        ]

    @property
    def num_materialized(self) -> int:
        return self.next_slot_idx

    def get_output_exprs(self) -> List[Expr]:
        """Returns exprs that were requested in the c'tor and require evaluation"""
        return self.output_exprs

    def _next_slot_idx(self) -> int:
        result = self.next_slot_idx
        self.next_slot_idx += 1
        return result

    def _record_unique_expr(self, expr: Expr, recursive: bool) -> Expr:
        """Records the expr if it's not a duplicate and assigns a slot idx to expr and its components"
        Returns:
            the unique expr
        """
        if expr in self.unique_exprs:
            # expr is a duplicate: we use the original instead
            return self.unique_exprs[expr]

        # expr value needs to be computed via Expr.eval()
        if recursive:
            for i, c in enumerate(expr.components):
                # make sure we only refer to components that have themselves been recorded
                expr.components[i] = self._record_unique_expr(c, True)
        assert expr.slot_idx < 0
        expr.slot_idx = self._next_slot_idx()
        self.unique_exprs.append(expr)
        return expr

    def _record_output_expr_id(self, e: Expr, output_expr_id: int) -> None:
        self.output_expr_ids[e.slot_idx].add(output_expr_id)
        for d in e.dependencies():
            self._record_output_expr_id(d, output_expr_id)

    def _compute_dependencies(self, target_slot_idxs: List[int], excluded_slot_idxs: List[int]) -> List[int]:
        """Compute exprs needed to materialize the given target slots, excluding 'excluded_slot_idxs'"""
        dependencies = [set() for _ in range(self.num_materialized)]  # indexed by slot_idx
        # doing this front-to-back ensures that we capture transitive dependencies
        max_target_slot_idx = max(target_slot_idxs)
        for expr in self.unique_exprs:
            if expr.slot_idx > max_target_slot_idx:
                # we're done
                break
            if expr.slot_idx in excluded_slot_idxs:
                continue
            if expr.slot_idx in self.input_expr_slot_idxs:
                # this is input and therefore doesn't depend on other exprs
                continue
            for d in expr.dependencies():
                if d.slot_idx in excluded_slot_idxs:
                    continue
                dependencies[expr.slot_idx].add(d.slot_idx)
                dependencies[expr.slot_idx].update(dependencies[d.slot_idx])
        # merge dependencies and convert to list
        return sorted(set().union(*[dependencies[i] for i in target_slot_idxs]))

    def substitute_exprs(self, expr_list: List[Expr], remove_duplicates: bool = True) -> None:
        """Substitutes exprs with their executable counterparts from unique_exprs and optionally removes duplicates"""
        i = 0
        unique_ids: Set[i] = set()  # slot idxs within expr_list
        while i < len(expr_list):
            unique_expr = self.unique_exprs[expr_list[i]]
            if unique_expr.slot_idx in unique_ids and remove_duplicates:
                del expr_list[i]
            else:
                expr_list[i] = unique_expr
                unique_ids.add(unique_expr.slot_idx)
                i += 1

    def get_dependencies(self, targets: List[Expr], exclude: Optional[List[Expr]] = None) -> List[Expr]:
        """
        Return list of dependencies needed to evaluate the given target exprs (expressed as slot idxs).
        The exprs given in 'exclude' are excluded.
        Returns:
            list of Exprs from unique_exprs (= with slot_idx set)
        """
        if exclude is None:
            exclude = []
        if len(targets) == 0:
            return []
        # make sure we only refer to recorded exprs
        targets = [self.unique_exprs[e] for e in targets]
        exclude = [self.unique_exprs[e] for e in exclude]
        target_slot_idxs = [e.slot_idx for e in targets]
        excluded_slot_idxs = [e.slot_idx for e in exclude]
        all_dependencies = set(self._compute_dependencies(target_slot_idxs, excluded_slot_idxs))
        all_dependencies.update(target_slot_idxs)
        result_ids = list(all_dependencies)
        result_ids.sort()
        return [self.unique_exprs[id] for id in result_ids]

    def create_eval_ctx(self, targets: List[Expr], exclude: Optional[List[Expr]] = None) -> EvalCtx:
        """Return EvalCtx for targets"""
        if exclude is None:
            exclude = []
        if len(targets) == 0:
            return self.EvalCtx([], [], [], [])
        dependencies = self.get_dependencies(targets, exclude)
        targets = [self.unique_exprs[e] for e in targets]
        target_slot_idxs = [e.slot_idx for e in targets]
        ctx_slot_idxs = [e.slot_idx for e in dependencies]
        return self.EvalCtx(
            slot_idxs=ctx_slot_idxs, exprs=[self.unique_exprs[slot_idx] for slot_idx in ctx_slot_idxs],
            target_slot_idxs=target_slot_idxs, target_exprs=targets)

    def set_exc(self, data_row: DataRow, slot_idx: int, exc: Exception) -> None:
        """Record an exception in data_row and propagate it to dependents"""
        data_row.set_exc(slot_idx, exc)
        for slot_idx in self.dependents[slot_idx]:
            data_row.set_exc(slot_idx, exc)

    def eval(
            self, data_row: DataRow, ctx: EvalCtx, profile: Optional[ExecProfile] = None, ignore_errors: bool = False
    ) -> None:
        """
        Populates the slots in data_row given in ctx.
        If an expr.eval() raises an exception, records the exception in the corresponding slot of data_row
        and omits any of that expr's dependents's eval().
        profile: if present, populated with execution time of each expr.eval() call; indexed by expr.slot_idx
        ignore_errors: if False, raises ExprEvalError if any expr.eval() raises an exception
        """
        for expr in ctx.exprs:
            assert expr.slot_idx >= 0
            if data_row.has_val[expr.slot_idx] or data_row.has_exc(expr.slot_idx):
                continue
            try:
                start_time = time.perf_counter()
                expr.eval(data_row, self)
                if profile is not None:
                    profile.eval_time[expr.slot_idx] += time.perf_counter() - start_time
                    profile.eval_count[expr.slot_idx] += 1
            except Exception as exc:
                _, _, exc_tb = sys.exc_info()
                self.set_exc(data_row, expr.slot_idx, exc)
                if not ignore_errors:
                    input_vals = [data_row[d.slot_idx] for d in expr.dependencies()]
                    raise excs.ExprEvalError(
                        expr, f'expression {expr}', data_row.get_exc(expr.slot_idx), exc_tb, input_vals, 0)

    def create_table_row(self, data_row: DataRow, exc_col_ids: Set[int]) -> Tuple[Dict[str, Any], int]:
        """Create a table row from the slots that have an output column assigned

        Return Tuple[dict that represents a stored row (can be passed to sql.insert()), # of exceptions]
            This excludes system columns.
        """
        num_excs = 0
        table_row: Dict[str, Any] = {}
        for info in self.table_columns:
            col, slot_idx = info.col, info.slot_idx
            if data_row.has_exc(slot_idx):
                # exceptions get stored in the errortype/-msg columns
                exc = data_row.get_exc(slot_idx)
                num_excs += 1
                exc_col_ids.add(col.id)
                table_row[col.storage_name()] = None
                table_row[col.errortype_storage_name()] = type(exc).__name__
                table_row[col.errormsg_storage_name()] = str(exc)
            else:
                val = data_row.get_stored_val(slot_idx)
                table_row[col.storage_name()] = val
                # we unfortunately need to set these, even if there are no errors
                table_row[col.errortype_storage_name()] = None
                table_row[col.errormsg_storage_name()] = None

        for slot_idx, col in enumerate(self.index_columns):
            if col is None:
                continue
            # don't use get_stored_val() here, we need to pass in the ndarray
            val = data_row[slot_idx]
            table_row[col.index_storage_name()] = val

        return table_row, num_excs




from typing import Optional

import pixeltable as pxt


@pxt.udf
def completions(
        prompt: str,
        model: str,
        max_tokens: Optional[int] = None,
        repetition_penalty: Optional[float] = None,
        stop: Optional[list] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        temperature: Optional[float] = None
) -> dict:
    import together
    return together.Complete.create(
        prompt,
        model,
        max_tokens=max_tokens,
        repetition_penalty=repetition_penalty,
        stop=stop,
        top_k=top_k,
        top_p=top_p,
        temperature=temperature
    )
